from dataclasses import dataclass

import numpy as np
import numpy.typing as npt

from deltakit_explorer.analysis.error_budget._discretisation import (
    DiscretisationStrategy,
)


@dataclass(frozen=True)
class FittingParameters:
    num_points_per_parameters: int = 10
    """Number of different values to try for each noise parameter.

    Corresponds to the number of points that will be used to fit a degree ``fitting_degree``
    polynomial. As such, should be greater than ``fitting_degree + 1``.
    """
    discretisation_strategy: DiscretisationStrategy = DiscretisationStrategy.LOGARITHMIC
    """A strategy to generate points that will be used to compute 1 / Λ on different
    values and fit a degree ``fitting_degree`` polynomial.

    Default to logarithmically spaced points.
    """
    fitting_degree: int = 3
    """Degree of polynomial that will be used to approximate 1 / Λ and to compute each
    of its derivatives.

    Should be lower than ``num_points_per_parameters - 1``. Higher values will incur
    higher standard deviation. Default to ``3``, which seems to be a good compromise
    between fit accuracy and resulting standard deviation.
    """

    def __post_init__(self) -> None:
        if self.num_points_per_parameters + 1 < self.fitting_degree + 2:
            msg = (
                f"Estimation of the standard deviation requires at least "
                f"fitting_degree + 2 = {self.fitting_degree + 2} discretisation points, "
                f"but only {self.num_points_per_parameters} + 1 are provided. Please "
                f"increase num_points_per_parameters to at least "
                f"{self.fitting_degree + 1}."
            )
            raise ValueError(msg)

    def get_discretisation(
        self, a: float, b: float, c: float
    ) -> npt.NDArray[np.floating]:
        return self.discretisation_strategy(
            a, b, c, self.num_points_per_parameters, self.fitting_degree
        )


@dataclass(frozen=True)
class SamplingParameters:
    max_shots: int = 10_000_000
    """Maximum number of shots per sampling task.

    A sampling task may stop with a lower number of samples if additional conditions are
    met, see ``lep_target_rse`` or ``lep_computation_min_fails`` for more details.
    """
    batch_size: int = 10_000
    """Number of sampling experiments that are submitted per batch."""
    lep_target_rse: float = 1e-4
    """Target relative standard error under which a sampling task is considered precise
    enough and can be stopped before ``max_shots`` sampling tasks have returned."""
    lep_computation_min_fails: int = 10
    """Minimum number of failures that should be witnessed before stopping a sampling task.

    A sampling task may stop with less failures, for example if ``max_shots`` shots have
    been performed."""
    max_workers: int = 1
    """Max number of parallel processes used by the function.

    Default to ``1`` which means fully sequential.
    """



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import deltakit_circuit
import numpy as np
import numpy.typing as npt
import stim
from deltakit_decode._abstract_matching_decoders import DecoderProtocol

if TYPE_CHECKING:
    from deltakit_core.decoding_graphs import OrderedDecodingEdges, OrderedSyndrome

from deltakit_explorer import Client, enums, types


class _CloudDecoder(DecoderProtocol):
    """Decoder (Cloud-based).

    Parameters
    ----------
    circuit : deltakit_circuit.Circuit | stim.Circuit | str
        Circuit to use to construct the decoder.
    parameters : Optional[dict[str, Any]]
        Some decoders may require additional parameters. Please refer
        to decoder documentation.
    use_experimental_graph_method : bool
        If enabled, the decoder tries to extract noise model from the
        experimental data.
    client : Optional[Client]
        Client with which to perform operations.

    Raises
    ------
    NotImplementedError
        When a method or attribute other than `decode_batch_to_logical_flip` is
        accessed.

    Notes
    -----
    Currently, only `decode_batch_to_logical_flip` is implemented; this class can be
    used with `deltakit.decode.analysis.run_decoding_on_circuit`, but not other
    decoder workflows.

    """

    _decoder_type: enums.DecoderType = enums.DecoderType.MWPM

    def __init__(
        self,
        circuit: deltakit_circuit.Circuit | stim.Circuit | str,
        parameters: dict[str, Any] | None = None,
        use_experimental_graph_method: bool = False,
        client: Client | None = None,
    ):
        # some leakage-aware decoders will need to hint this value,
        # if they are using non-standard STIM extension
        self.num_observables = 0
        self.decoder_parameters = {} if parameters is None else parameters
        self.use_experimental_graph = use_experimental_graph_method
        self.stim_circuit: stim.Circuit | None = None
        # to communicate to the server and to support leakage,
        # circuit must be a string
        if isinstance(circuit, deltakit_circuit.Circuit):
            circuit = circuit.as_stim_circuit()
        if isinstance(circuit, stim.Circuit):
            self.stim_circuit = circuit
            self.num_observables = self.stim_circuit.num_observables
            circuit = str(circuit)
        self.text_circuit = circuit
        if client is None:
            msg = "Currently, a `client` must be provided to instantiate this class."
            raise NotImplementedError(msg)
        self.client = client

    def decode_batch_to_logical_flip(
            self,
            syndrome_batch: npt.NDArray[np.uint8],
            leakage_batch: npt.NDArray[np.uint8] | None = None,
        ):
        """The method decodes the batch of syndromes to boolean values."""
        num_shots = syndrome_batch.shape[0]
        detectors = types.DetectionEvents(syndrome_batch)
        if self.num_observables < 1:
            msg = (
                "Circuit must have at least one observable. "
                "Please make sure your circuit has observables or provide "
                f"`num_observables` when instantiating an `{self.__class__.__name__}`."
            )
            raise ValueError(msg)
        observables = types.ObservableFlips(
            np.zeros((num_shots, self.num_observables), dtype=syndrome_batch.dtype)
        )
        leakage = None
        if leakage_batch is not None:
            leakage = types.LeakageFlags(leakage_batch)
        decoder = types.Decoder(
            decoder_type=self._decoder_type,
            use_experimental_graph=self.use_experimental_graph,
            parameters=self.decoder_parameters
        )
        decoding_result = self.client.decode(
            detectors=detectors,
            observables=observables,
            decoder=decoder,
            noisy_stim_circuit=self.text_circuit,
            leakage_flags=leakage,
        )
        return decoding_result.predictions.as_numpy()


    def decode_batch_to_full_correction(
        self, syndrome_batch: npt.NDArray[np.uint8]
    ) -> npt.NDArray[np.uint8]:
        raise NotImplementedError()

    def decode_to_logical_flip(self, syndrome: OrderedSyndrome) -> tuple[bool, ...]:
        raise NotImplementedError()

    def decode_to_full_correction(
        self, syndrome: OrderedSyndrome
    ) -> OrderedDecodingEdges:
        raise NotImplementedError()

    @property
    def logicals_edge_list(self) -> list[list[int]]:
        raise NotImplementedError()


class MWPMDecoder(_CloudDecoder):
    """Minimum Weight Perfect Matching Decoder (Cloud-based).

    Parameters
    ----------
    circuit : deltakit_circuit.Circuit | stim.Circuit | str
        Circuit to use to construct the decoder.
    parameters : Optional[dict[str, Any]]
        Not used in MWPM.
    use_experimental_graph_method : bool
        If enabled, the decoder tries to extract noise model from the
        experimental data.
    client : Optional[Client]
        Client with which to perform operations.

    Raises
    ------
    NotImplementedError
        When a method or attribute other than `decode_batch_to_logical_flip` is
        accessed.

    Notes
    -----
    Currently, only `decode_batch_to_logical_flip` is implemented; this class can be
    used with `deltakit.decode.analysis.run_decoding_on_circuit`, but not other
    decoder workflows.

    This cloud-based decoder implements Minimum Weight Perfect Matching
    (https://arxiv.org/abs/2303.15933).
    """
    _decoder_type = enums.DecoderType.MWPM


class CCDecoder(_CloudDecoder):
    """Collision Clustering Decoder (Cloud-based).

    Parameters
    ----------
    circuit : deltakit_circuit.Circuit | stim.Circuit | str
        Circuit to use to construct the decoder.
    parameters : Optional[dict[str, Any]]
        Not used in CC.
    use_experimental_graph_method : bool
        If enabled, the decoder tries to extract noise model from the
        experimental data.
    client : Optional[Client]
        Client with which to perform operations.

    Raises
    ------
    NotImplementedError
        When a method or attribute other than `decode_batch_to_logical_flip` is
        accessed.

    Notes
    -----
    Currently, only `decode_batch_to_logical_flip` is implemented; this class can be
    used with `deltakit.decode.analysis.run_decoding_on_circuit`, but not other
    decoder workflows.

    This cloud-based decoder implements Collision Clustering
    (https://arxiv.org/abs/2309.05558).
    """
    _decoder_type = enums.DecoderType.CC


class BeliefMatchingDecoder(_CloudDecoder):
    """Belief Matching Decoder (Cloud-based).

    Parameters
    ----------
    circuit : deltakit_circuit.Circuit | stim.Circuit | str
        Circuit to use to construct the decoder.
    parameters : Optional[dict[str, Any]]
        Not used in Belief Matching.
    use_experimental_graph_method : bool
        If enabled, the decoder tries to extract noise model from the
        experimental data.
    client : Optional[Client]
        Client with which to perform operations.

    Raises
    ------
    NotImplementedError
        When a method or attribute other than `decode_batch_to_logical_flip` is
        accessed.

    Notes
    -----
    Currently, only `decode_batch_to_logical_flip` is implemented; this class can be
    used with `deltakit.decode.analysis.run_decoding_on_circuit`, but not other
    decoder workflows.

    This cloud-based decoder implements Belief Matching
    (https://journals.aps.org/prx/abstract/10.1103/PhysRevX.13.031007).
    """
    _decoder_type = enums.DecoderType.BELIEF_MATCHING


class BPOSDecoder(_CloudDecoder):
    """Belief Propagation - Ordered Statistics Decoder (Cloud-based).

    Parameters
    ----------
    circuit : deltakit_circuit.Circuit | stim.Circuit | str
        Circuit to use to construct the decoder.
    parameters : Optional[dict[str, Any]]
        - `max_bp_rounds` is an integer, and specifies the maximum number
          of iterations of message passing that should be performed during
          the execution of belief propagation.
          It may terminate earlier. By default, this is 20.
        - `combination_sweep_order` is the depth of the OSD search.
    use_experimental_graph_method : bool
        If enabled, the decoder tries to extract noise model from the
        experimental data.
    client : Optional[Client]
        Client with which to perform operations.

    Raises
    ------
    NotImplementedError
        When a method or attribute other than `decode_batch_to_logical_flip` is
        accessed.

    Notes
    -----
    Currently, only `decode_batch_to_logical_flip` is implemented; this class can be
    used with `deltakit.decode.analysis.run_decoding_on_circuit`, but not other
    decoder workflows.

    This cloud-based decoder implements Belief Propagation - Ordered Statistics Decoding
    (https://quantum-journal.org/papers/q-2021-11-22-585/).
    """
    _decoder_type = enums.DecoderType.BP_OSD


class ACDecoder(_CloudDecoder):
    """Ambiguity Clustering Decoder (Cloud-based).

    Parameters
    ----------
    circuit : deltakit_circuit.Circuit | stim.Circuit | str
        Circuit to use to construct the decoder.
    parameters : Optional[dict[str, Any]]
        - `bp_rounds` is an integer, and specifies how many iterations
          of message passing should be performed during the execution
          of belief propagation. Note that bp_rounds in AC is different
          `from max_bp_rounds` in BP_OSD as early termination is not
          allowed. Typically, setting this equal to the distance of
          the code is sufficient. By default, this is 20.
        - `ac_kappa_proportion` is a float, between 0.0 and 1.0, and
          reflects the number of error mechanisms, in addition to those
          used to find a first solution, that should be used to grow
          clusters to search for additional solutions, expressed as a
          proportion of the total number of error mechanisms. Setting
          this number higher results in better accuracy at the cost of slower
          performance. Start with 0.0 and increase by 0.01 until the desired
          accuracy is reached. Reasonable values lie between 0 and 0.1, as
          larger values will typically lead to a significant slow down.
          By default, this is 0.01.
    use_experimental_graph_method : bool
        If enabled, the decoder tries to extract noise model from the
        experimental data.
    client : Optional[Client]
        Client with which to perform operations.

    Raises
    ------
    NotImplementedError
        When a method or attribute other than `decode_batch_to_logical_flip` is
        accessed.

    Notes
    -----
    Currently, only `decode_batch_to_logical_flip` is implemented; this class can be
    used with `deltakit.decode.analysis.run_decoding_on_circuit`, but not other
    decoder workflows.

    This cloud-based decoder implements Ambiguity Clustering
    (https://arxiv.org/abs/2406.14527).
    """
    _decoder_type = enums.DecoderType.AC


class LCDecoder(_CloudDecoder):
    """Local Clustering Decoder (Cloud-based).

    Parameters
    ----------
    circuit : deltakit_circuit.Circuit | stim.Circuit | str
        Circuit to use to construct the decoder.
    parameters : Optional[dict[str, Any]]
        Some decoders may require additional parameters. Please refer
        to decoder documentation.
    use_experimental_graph_method : bool
        If enabled, the decoder tries to extract noise model from the
        experimental data.
    client : Optional[Client]
        Client with which to perform operations.
    num_observables : Optional[int]
        If provided, sets the number of observables in the decoder.
        If not provided, the number of observables is inferred from
        the circuit (if possible). This parameter is useful when
        the circuit uses non-standard STIM extensions to represent
        observables.

    Raises
    ------
    NotImplementedError
        When a method or attribute other than `decode_batch_to_logical_flip` is
        accessed.

    Notes
    -----
    Currently, only `decode_batch_to_logical_flip` is implemented; this class can be
    used with `deltakit.decode.analysis.run_decoding_on_circuit`, but not other
    decoder workflows.

    This cloud-based decoder implements the Local Clustering Decoder algorithm
    (https://arxiv.org/abs/2411.10343).
    """
    _decoder_type = enums.DecoderType.LCD


    def __init__(
        self,
        circuit: deltakit_circuit.Circuit | stim.Circuit | str,
        parameters: dict[str, Any] | None = None,
        use_experimental_graph_method: bool = False,
        client: Client | None = None,
        num_observables: int | None = None,
    ):
        """Local Clustering Decoder (Cloud-based).

        Parameters
        ----------
        circuit : deltakit_circuit.Circuit | stim.Circuit | str
            Circuit to use to construct the decoder.
        parameters : Optional[dict[str, Any]]
            Some decoders may require additional parameters. Please refer
            to decoder documentation.
        use_experimental_graph_method : bool
            If enabled, the decoder tries to extract noise model from the
            experimental data.
        client : Optional[Client]
            Client with which to perform operations.
        num_observables : Optional[int]
            If provided, sets the number of observables in the decoder.
            If not provided, the number of observables is inferred from
            the circuit (if possible). This parameter is useful when
            the circuit uses non-standard STIM extensions to represent
            observables.

        """
        super().__init__(
            circuit,
            parameters,
            use_experimental_graph_method,
            client,
        )
        if num_observables is not None:
            self.num_observables = num_observables



import math
from collections.abc import Callable, Iterator, Mapping, Sequence

import numpy as np
import numpy.typing as npt
import pandas as pd
from deltakit_circuit._circuit import Circuit
from deltakit_decode.analysis._run_all_analysis_engine import RunAllAnalysisEngine

from deltakit_explorer.analysis.error_budget._generation import (
    generate_decoder_managers_for_lambda,
)
from deltakit_explorer.analysis.error_budget._memory import (
    MemoryGenerator,
    PreComputedMemoryGenerator,
    get_rotated_surface_code_memory_circuit,
)
from deltakit_explorer.analysis.error_budget._parameters import (
    FittingParameters,
    SamplingParameters,
)
from deltakit_explorer.analysis.error_budget._post_processing import (
    compute_lambda_and_stddev_from_results,
)


def _variate_ith_parameter_by(
    central_point: npt.NDArray[np.floating],
    variations: npt.NDArray[np.floating],
    i: int,
) -> Iterator[npt.NDArray[np.floating]]:
    """Returns versions of ``central_point`` where the ``i``-th parameter is
    successively replaced by values in ``variations``.

    Args:
        central_point (npt.NDArray[numpy.floating]): base 1-dimensional array of numbers
            of shape ``(n,)`` that will be copied, modified on the ``i``-th variable and
            returned.
        variations (npt.NDArray[numpy.floating]): 1-dimensional array of shape ``(m,)``
            containing values that should be used to replace the ``i``-th entry of
            ``central_point``.
        i (int): index of the entry in ``central_point`` that should be changed.

    Yields:
        ``m`` arrays of shape ``(n,)`` that are copies of ``central_point`` with the
        ``i``-th coordinate entry replaced with an entry from ``variations``.
    """
    central_point = central_point.reshape((-1, 1))
    variations = variations.reshape((-1,))
    parameters = np.tile(central_point, (1, variations.size))
    parameters[i, :] = variations
    yield from parameters.T


def _approximate_derivative_at_point_from_values(
    x: npt.NDArray[np.floating],
    y: npt.NDArray[np.floating],
    stddevs: npt.NDArray[np.floating],
    gradient_approximation_point: float,
    degree: int = 3,
) -> tuple[float, float]:
    """Approximate the gradient at ``gradient_approximation_point`` from the given ``x``
    and ``y``.

    This function fits a degree ``degree`` polynomial to the points given by ``x``,
    ``y`` and ``stddevs`` (the standard deviation of each point in ``y``) and then
    computes the gradient of the fitted polynomial at ``gradient_approximation_points``.

    This algorithm is used to use as much as possible the standard deviation information
    and to avoid non-linear behaviour at the extremities of the interval containing all
    values in ``x`` to affect the gradient too much.

    Warning:
        This function will work better if ``gradient_approximation_point`` is close to
        the "center" of the interval formed by ``x``.

        This is due to the fact that the gradient at ``gradient_approximation_point`` is
        estimated through a polynomial of degree ``degree`` fitted on the provided
        ``x``, ``y`` and ``stddevs``, which inherently suffer from the Runge's
        phenomenon (even with the optimal fitting points for ``x``, see
        https://en.wikipedia.org/wiki/Runge%27s_phenomenon). In this case, Runge's
        phenomenon means that the furthest ``gradient_approximation_point`` will be from
        the "center" of the discretisation interval provided by ``x``, the more likely
        our gradient estimation is impacted significantly by the oscillations.

        It is considered to be the responsibility of the caller to check that the
        provided ``x`` and ``gradient_approximation_point`` are picked such that Runge's
        phenomenon will not impact our estimation too much.

    Args:
        x (npt.NDArray[numpy.floating]): exact values on which we evaluated a noisy
            function. Should be a 1-dimensional array.
        y (npt.NDArray[numpy.floating]): best estimation of the result obtained from the
            noisy function evaluation when evaluated on the corresponding entry in
            ``x``. Should be a 1-dimensional array.
        stddevs (npt.NDArray[numpy.floating]): standard deviation of the estimate in
            ``y``. Should be a 1-dimensional array.
        gradient_approximation_point (float): point at which the gradient should be
            estimated.
        degree (int): degree of the polynomial to fit the provided points and estimate
            the gradient at ``gradient_approximation_point``.

    Returns:
        value of the gradient (a single float because ``x``, ``y`` and ``stddevs`` are
        one-dimensional arrays) and the standard deviation of the estimate.
    """
    # Perform the approximation using the provided standard deviations
    coefficients, cov = np.polyfit(x, y, deg=degree, cov="unscaled", w=1 / stddevs)

    # Flipping the coefficients and covariance matrix to have the index corresponding to
    # the degree (i.e., ``coefficients[i]`` is multiplying ``x**i`` in the polynomial
    # and ``cov[i,i]`` is the variance of ``coefficients[i]``).
    coefficients, cov = np.flip(coefficients), np.flip(cov)

    # Compute the derivative
    derivative = float(
        sum(
            coefficient * (power + 1) * gradient_approximation_point**power
            for power, coefficient in enumerate(coefficients[1:])
        )
    )
    # Compute the variance of the derivative estimate
    standard_deviation = math.sqrt(
        _get_variance_of_gradient_estimation_at_point(cov, gradient_approximation_point)
    )
    return derivative, standard_deviation


def _get_variance_of_gradient_estimation_at_point(
    cov: npt.NDArray[np.floating], c: float
) -> float:
    """Get the variance of the gradient estimation at the point ``c`` for a polynomial
    with uncertainties on its coefficients provided by the covariance matrix ``cov``.

    Args:
        cov (npt.NDArray[numpy.floating]): an array of shape ``(d + 1, d + 1)``
            representing the covariance matrix of the coefficients defining the degree-d
            polynomial used to estimate the gradient.
        c (float): point at which the degree-d polynomial will be used to estimate the
            gradient value.

    Returns:
        The variance of the gradient estimation at point ``c``.
    """
    # From https://en.wikipedia.org/wiki/Covariance#Covariance_of_linear_combinations we
    # have an easy formula for the variance involving the covariance matrix. We build
    # the terms of the sum in a copy of the ``cov`` matrix.
    # We do not need to use the first row/column of the covariance matrix because it is
    # linked with the intersect, that is not used when computing the derivative.
    derivative_cov = np.copy(cov[1:, 1:])
    num_derivative_coefficients = derivative_cov.shape[0]
    # The original polynomial at ``x`` is given by
    #       sum(a[i] * x**i for i in range(0, n+1))
    # Its derivative is then given by
    #       sum(a[i+1] * (i+1) * x**i for i in range(0, n))
    # Removing a[0] from a (and so shifting its indexing by 1, which is what we just did
    # with the covariance matrix above) we have the derivative given by
    #       sum(a[i] * (i+1) * x**i for i in range(0, n))
    # When evaluating the gradient at the point c, the degree i coefficient of the
    # derivative (that is ``a[i]`` above or equivalently the degree (i+1) coefficient of
    # the original polynomial) is multiplied by (i + 1) * c**i. We can ignore (i.e., not
    # change) the degree 0 as the factor is (0 + 1) * c**0 = 1.
    for i in range(1, num_derivative_coefficients):
        derivative_cov[i, :] *= (i + 1) * c**i
        derivative_cov[:, i] *= (i + 1) * c**i
    # Finally, according to
    # https://en.wikipedia.org/wiki/Covariance#Covariance_of_linear_combinations,
    # the variance is given by the sum of the derivative covariance matrix above.
    return float(np.sum(derivative_cov))


def generate_sweep_parameters(
    central_point: npt.NDArray[np.floating],
    noise_parameters_exploration_bounds: list[tuple[float, float]],
    fitting_parameters: FittingParameters = FittingParameters(),
) -> npt.NDArray[np.floating]:
    """Generate multiple copies of the provided ``central_point`` with one parameter changed.

    Each parameter is changed according to the fitting strategy. The resulting array will contain
    copies of ``central_point`` where, for each copy, only a single entry is different from
    ``central_point``. The value of that entry, and the number of variations, depends on the
    provided ``fitting_parameters``.

    Args:
        central_point: point that should be copied and modified in the resulting array.
        noise_parameters_exploration_bounds: bounds for each parameters, that will impact the values
            that are inserted in the returned array.
        fitting_parameters: parameters of the fitting task that will control how many variations are
            inserted for each parameters and how the values of these variations will be obtained.

    Returns:
        An array of shape
        ``(fitting_parameters.num_points_per_parameters * central_point.size, central_point.size)``
        that contains as rows variations of ``central_point`` where a single parameter has been
        changed each time.
    """
    # Getting the points on which we will estimate 1 / Λ into ``noise_parameters``.
    # This is performing a sweeping for each parameter individually.
    xis: list[npt.NDArray[np.floating]] = [central_point.reshape((-1,))]
    for i, (minimum, maximum) in enumerate(noise_parameters_exploration_bounds):
        variations = fitting_parameters.get_discretisation(
            minimum, maximum, central_point[i]
        )
        xis.extend(_variate_ith_parameter_by(central_point, variations, i))

    return np.asarray(xis).T


def get_decoding_result(
    noise_model: Callable[[Circuit, npt.NDArray[np.floating]], Circuit],
    sweep_noise_parameters: npt.NDArray[np.floating],
    noise_parameter_names: list[str],
    num_rounds_by_distances: Mapping[int, Sequence[int]],
    fitting_parameters: FittingParameters = FittingParameters(),
    sampling_parameters: SamplingParameters = SamplingParameters(),
    memory_generator: MemoryGenerator = get_rotated_surface_code_memory_circuit,
) -> pd.DataFrame:
    """Construct, sample and decode the experiments represented by the input parameters
    and returns statistics on the success rate of each experiment.

    Experiments to run are given by the provided ``noise_model``,
    ``sweep_noise_parameters``, ``num_rounds_by_distances``, ``memory_generator``.

    Args:
        noise_model: noise model to use to annotate noise instructions on the circuits.
        sweep_noise_parameters: parameters to use for the noise model. Should be the array
            returned when calling ``generate_sweep_parameters``.
        noise_parameter_names: identifiers for each noise parameter, used to identify
            results in the returned statistics.
        num_rounds_by_distances (Mapping[int, Sequence[int]]): a mapping from each code
            distance that should be tested to the number of rounds that should be
            sampled in order to estimate the logical error-probability per round, to
            ultimately get 1 / Λ.
        fitting_parameters: additional parameters relating to how the gradient is
            estimated.
        sampling_parameters: additional parameters relating to the sampling tasks used to
            estimate 1 / Λ indirectly.
        memory_generator: a callable that can generate a memory experiment. The resulting
            circuit will go through the provided ``noise_model`` for different values of
            the noise parameters.

    Returns:
        A pandas DataFrame containing the statistics resulting from sampling the provided
        experiments.
    """
    # ``noise_parameters`` contains all the noise parameters we want to evaluate 1 / Λ.
    # Prepare the computation by building the decoder managers.
    decoder_managers = generate_decoder_managers_for_lambda(
        sweep_noise_parameters,
        noise_model,
        num_rounds_by_distances,
        sampling_parameters.max_workers,
        memory_generator=memory_generator,
        noise_parameter_names=noise_parameter_names,
    )

    # Start the computation
    num_parameters = sweep_noise_parameters.shape[0]
    num_points = num_parameters * fitting_parameters.num_points_per_parameters
    engine = RunAllAnalysisEngine(
        experiment_name=f"Estimating Λ on {num_points} points",
        decoder_managers=decoder_managers,
        max_shots=sampling_parameters.max_shots,
        batch_size=sampling_parameters.batch_size,
        # Early stopping when we have a low-enough standard deviation
        loop_condition=RunAllAnalysisEngine.loop_until_observable_rse_below_threshold(
            sampling_parameters.lep_target_rse,
            sampling_parameters.lep_computation_min_fails,
        ),
        num_parallel_processes=sampling_parameters.max_workers,
    )
    return engine.run()


def get_lambda_gradient(
    report: pd.DataFrame,
    central_point: npt.NDArray[np.floating],
    sweep_noise_parameters: npt.NDArray[np.floating],
    noise_parameter_names: list[str],
    num_rounds_by_distances: Mapping[int, Sequence[int]],
    fitting_parameters: FittingParameters = FittingParameters(),
) -> tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]]:
    """Post-process the provided ``report`` to compute an estimation of the gradient of 1 / Λ.

    Args:
        report: statistics returned by ``get_decoding_results``.
        central_point: point at which the gradient of 1 / Λ should be estimated.
        sweep_noise_parameters: parameters to use for the noise model. Should be the array
            returned when calling ``generate_sweep_parameters``.
        noise_parameter_names: identifiers for each noise parameter, used to identify
            results in the returned statistics.
        num_rounds_by_distances (Mapping[int, Sequence[int]]): a mapping from each code
            distance that should be tested to the number of rounds that should be
            sampled in order to estimate the logical error-probability per round, to
            ultimately get 1 / Λ.
        fitting_parameters: additional parameters relating to how the gradient is
            estimated.

    Returns:
        a tuple ``(gradient, stddev)`` containing the estimation of the gradient at the
        provided ``central_point`` and the standard deviation of the estimation.
    """
    # Post-process the results to get all the estimations for 1 / Λ
    lambdas, lambda_stddevs = compute_lambda_and_stddev_from_results(
        sweep_noise_parameters, noise_parameter_names, num_rounds_by_distances, report
    )
    lambda_reciprocals = 1 / lambdas
    lambda_reciprocal_stddevs = np.abs(lambda_stddevs / lambdas**2)

    # We now have all the estimations of 1 / Λ, we can approximate the gradient
    # Note that ``noise_parameters``, ``lambda_reciprocals`` and
    # ``lambda_reciprocal_stddevs`` have the same shapes: a 2-dimensional array with
    # values as columns. Additionally, the first column corresponds to the exact point
    # at which we want the gradient and each following group of
    # ``num_points_per_parameters`` columns correspond to the variation of one
    # parameter.
    gradient: list[float] = []
    gradient_stddev: list[float] = []
    for npi, noise_parameter in enumerate(central_point):
        start = 1 + fitting_parameters.num_points_per_parameters * npi
        end = 1 + fitting_parameters.num_points_per_parameters * (npi + 1)
        # Index 0 is ``central_point``, so it can be included in all estimations.
        column_indices = [0, *list(range(start, end))]
        x = sweep_noise_parameters[npi, column_indices]
        y = lambda_reciprocals[0, column_indices]
        stddevs = lambda_reciprocal_stddevs[0, column_indices]
        derivative, derivative_stddev = _approximate_derivative_at_point_from_values(
            x, y, stddevs, noise_parameter, degree=fitting_parameters.fitting_degree
        )
        gradient.append(derivative)
        gradient_stddev.append(derivative_stddev)
    return np.asarray(gradient), np.asarray(gradient_stddev)


def inverse_lambda_gradient_at(
    noise_model: Callable[[Circuit, npt.NDArray[np.floating]], Circuit],
    noise_parameters: npt.NDArray[np.floating] | Sequence[float],
    num_rounds_by_distances: Mapping[int, Sequence[int]],
    noise_parameters_exploration_bounds: list[tuple[float, float]],
    fitting_parameters: FittingParameters = FittingParameters(),
    sampling_parameters: SamplingParameters = SamplingParameters(),
    memory_generator: MemoryGenerator
    | Mapping[int, Mapping[int, Circuit]] = get_rotated_surface_code_memory_circuit,
) -> tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]]:
    """The gradient of 1 / Λ at the provided ``noise_model_parameters``.

    Args:
        noise_model (Callable[[Circuit, npt.NDArray[np.floating]], Circuit]): a callable
            adding noise to the provided circuit, according to the parameters provided.
        noise_parameters (npt.NDArray[numpy.floating] | Sequence[float]): valid
            parameters to forward to ``noise_model`` representing the point at which the
            gradient should be computed.
        num_rounds_by_distances (Mapping[int, Sequence[int]]): a mapping from each code
            distance that should be tested to the number of rounds that should be
            sampled in order to estimate the logical error-probability per round, to
            ultimately get 1 / Λ.
        noise_parameters_exploration_bounds (list[tuple[float, float]]): ``(min, max)``
            bounds for each noise parameter of the provided ``noise_model``. A degree
            ``fitting_degree`` polynomial will be fitted on the interval ``[min, max]``.
            The corresponding noise parameter from the provided ``noise_model`` should
            be strictly contained in ``[min, max]`` (i.e., for any valid ``i``, the
            following is true:
            ``noise_parameters_exploration_bounds[i][0] <
            noise_model.noise_parameters[i] <
            noise_parameters_exploration_bounds[i][1]``). Ideally, the lower (resp.
            upper) bound provided must be such that the logical error probability when
            replacing the parameter with its lower (resp. upper) bound is above
            ``100 / max_shots`` to ensure enough fails are observed with ``max_shots``
            shots (resp. below ``1 / 2`` to ensure that we can compute the logical error
            probability per round).
        fitting_parameters: additional parameters relating to how the gradient is
            estimated.
        sampling_parameters: additional parameters relating to the sampling tasks used to
            estimate 1 / Λ indirectly.
        memory_generator (MemoryGenerator): a callable that can generate a memory
            experiment. The resulting circuit will go through the provided
            ``noise_model`` for different values of the noise parameters.


    Returns:
        the error-budgeting result, which consists of an array of contributions for each
        of the noise parameters of the provided ``noise_model`` along with their
        associated standard deviations. Can also include the estimated value of Λ on the
        provided noise model parameter if ``include_lambda`` is ``True``.
    """

    # Making sure that the memory generator is an object implementing the MemoryGenerator
    # protocol.
    if isinstance(memory_generator, Mapping):
        memory_generator = PreComputedMemoryGenerator(memory_generator)

    # Make sure that noise_model_parameters is a numpy array, even if a generic Sequence
    # is provided, as this is simpler for later.
    noise_model_parameters = np.asarray(noise_parameters)
    # Create unique identifiers for noise parameters that will be used to discriminate between them
    # in the CSV file storing the simulation results.
    noise_parameter_names = [str(i) for i in range(len(noise_parameters))]

    central_point: npt.NDArray[np.floating] = noise_model_parameters.reshape((-1, 1))
    # Generate the noise parameters at which we will compute Λ.
    sweep_noise_parameters = generate_sweep_parameters(
        central_point, noise_parameters_exploration_bounds, fitting_parameters
    )
    # Sample on these noise parameters.
    report = get_decoding_result(
        noise_model,
        sweep_noise_parameters,
        noise_parameter_names,
        num_rounds_by_distances,
        fitting_parameters,
        sampling_parameters,
        memory_generator,
    )
    # Compute the gradient of Λ from the sampling results.
    return get_lambda_gradient(
        report,
        central_point,
        sweep_noise_parameters,
        noise_parameter_names,
        num_rounds_by_distances,
        fitting_parameters,
    )



"""Matplotlib visualisation of error budgets."""

from __future__ import annotations

import colorsys
import warnings
from collections.abc import Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Any

import matplotlib.pyplot as plt
import numpy as np
from adjustText import adjust_text
from matplotlib.colors import ColorConverter, to_rgb

if TYPE_CHECKING:
    import numpy.typing as npt
    from matplotlib.axes import Axes
    from matplotlib.figure import Figure
    from matplotlib.text import Text

_RIVERLANE_PLOT_COLOURS: list[str] = [
    "#006F62",
    "#FF6A00",
    "#4B5BFF",
    "#E01500",
    "#003B40",
    "#D835A2",
    "#00A3A3",
    "#9B3EC6",
    "#54170E",
    "#FFD00B",
    "#0AD2F2",
]
_RIVERLANE_DARK_TEXT_COLOUR = "#0A1600"
_RIVERLANE_WHITE_TEXT_COLOUR = "#E5F0EF"
_RIVERLANE_LINE_COLOUR = "#003B40"
_RIVERLANE_NEUTRAL_GREY = "#DBDBDB"


def _scale_lightness(rgb: tuple[float, float, float] | str, scale_l: float):
    # From https://stackoverflow.com/a/60562502
    if isinstance(rgb, str):
        rgb = ColorConverter.to_rgb(rgb)
    # convert rgb to hls
    # Ignore E741 because "l" here is not really ambiguous and this function is small.
    h, l, s = colorsys.rgb_to_hls(*rgb)  # noqa: E741
    # manipulate h, l, s values and return as rgb
    return colorsys.hls_to_rgb(h, min(1, l * scale_l), s=s)


def _colour_brightness(colour: str) -> float:
    r, g, b = to_rgb(colour)
    return 0.2125 * r + 0.7152 * g + 0.0722 * b


def _text_size_in_data_coordinates(
    fig: Figure, ax: Axes, text: str
) -> tuple[float, float]:
    """Get the size of ``text`` when drawn on ``fig``.

    Adapted From https://stackoverflow.com/a/36959454.
    """
    get_renderer = getattr(fig.canvas, "get_renderer")
    assert get_renderer is not None
    renderer = get_renderer()
    text_object = ax.text(0.5, 0.5, text)
    bb = text_object.get_window_extent(renderer=renderer).transformed(
        ax.transData.inverted()
    )
    text_object.remove()
    return (bb.width, bb.height)


def _draw_line_from_text_to_position(
    fig: Figure, ax: Axes, text: Text, x: float, y: float, colour: str, t: float = 0.2
) -> None:
    """Draw a line between text and (x, y) but with the end at (x, y) slightly shortened."""
    # First, get the coordinates at which the line will intersect with the bounding-box
    # of the text.
    get_renderer = getattr(fig.canvas, "get_renderer")
    assert get_renderer is not None
    renderer = get_renderer()
    bl, tr = ax.transData.inverted().transform(text.get_window_extent(renderer))
    cx, cy = (tr + bl) / 2
    slope = (cy - y) / (cx - x)
    offset = cy - slope * cx
    x_intersect, y_intersect = (bl[1] - offset) / slope, bl[1]

    x, y = (1 - t) * x + t * x_intersect, (1 - t) * y + t * y_intersect
    ax.plot([x, x_intersect], [y, y_intersect], color=colour)


def plot_error_budget(
    contributions: npt.NDArray[np.floating] | Sequence[float],
    contribution_descriptions: Sequence[str],
    contributions_stddev: (
        npt.NDArray[np.floating] | Sequence[float | None] | None
    ) = None,
    fig: Figure | None = None,
    ax: Axes | None = None,
    save_to_filename: Path | None = None,
    excess_noise_description: str = "excess",
) -> tuple[Figure, Axes]:
    """Plot the provided error budget.

    Args:
        contributions (npt.NDArray[numpy.floating] | Sequence[float]):
            error-budget as an array of numbers.
        contribution_descriptions (Sequence[str]): textual description of each error
            mechanism involved in ``contributions``. Will be plotted on the error-budget
            figure.
        contributions_stddev (npt.NDArray[numpy.floating] | Sequence[float | None] | None):
            standard deviation of the provided values in ``contributions``. Default to
            ``None``, which does not draw the error-bars. If an array is provided,
            ``None`` values mean "no error-bars for this value" (this is often the case
            for the excess noise).
        fig (Figure | None): the figure to draw the error budget on. If not provided,
            a new figure and ax will be generated by calling ``plt.subplots()``.
        ax (Axes | None): the ax to draw the error budget on. If not provided,
            a new figure and ax will be generated by calling ``plt.subplots()``.
        save_to_filename (Path | None): if provided, the drawn figure will be saved to
            this path before returning.
        excess_noise_description (str): description of the budget that is associated
            with excess noise that has not been modelised in the error-budgeting
            process. If it matches (case-insensitive) one or more of the entries in
            ``contribution_descriptions``, this entry will be draw with grey hatches.

    Returns:
        Tuple[Figure, Axes]:
            The figure and ax that have been used to draw the resulting error-budget
            plot.

    Examples:
        Calculating per-round logical error probability and its standard deviation
        given number of fails, and number of shots for several rounds::

            import matplotlib.pyplot as plt

            fig, ax = plot_error_budget(
                contributions=[1.9, 0.3, 0.9, 0.6, 0.15, 0.4, 0.65],
                contribution_descriptions=[
                    "CZ",
                    "CZ stray int.",
                    "Data idle",
                    "Meas.",
                    "Leakage",
                    "1Q",
                    "Excess",
                ],
                excess_noise_description="Excess",
            )
            plt.show()
    """
    # Declare dimensions here. Every dimension is defined as a fraction of bar_height to
    # have scaling independent plots.
    bar_height: float = 0.5
    y_axis_offset_from_zero: float = bar_height / 10
    bar_center: float = y_axis_offset_from_zero + bar_height / 2
    bar_top: float = y_axis_offset_from_zero + bar_height
    non_fitting_texts_ys: list[float] = [bar_top + i * bar_height / 3 for i in (1, 2)]
    ymax = non_fitting_texts_ys[-1] + bar_height / 2

    # In case the user did not provide a figure or an axe, get it ourselves.
    if fig is None or ax is None:
        fig, ax = plt.subplots()

    # Plot a stacked horizontal bar plot. Do not include labels that do not fit in their
    # respective bar, but keep them to include them later.
    non_fitting_texts: list[str] = []
    nft_centres: list[float] = []
    nft_bar_colours: list[str] = []
    nft_bar_centres: list[float] = []
    twidths: list[float] = [
        _text_size_in_data_coordinates(fig, ax, d)[0] for d in contribution_descriptions
    ]

    offset: float = 0
    for i in range(len(contributions)):
        colour = _RIVERLANE_PLOT_COLOURS[i % len(_RIVERLANE_PLOT_COLOURS)]
        description = contribution_descriptions[i]
        contribution = contributions[i]
        twidth = twidths[i]
        # We need to handle the special case of the excess noise. Doing that with
        # kwargs.
        kwargs: dict[str, Any] = {
            "color": colour,
            "edgecolor": colour,
            "ecolor": _scale_lightness(colour, 0.8),
        }
        if description.lower() == excess_noise_description.lower():
            # Note that the RGB values of the colour below matter: we want it to be
            # white because the background will be white, and text colour will be picked
            # according to the background colour brightness.
            kwargs["color"] = "#FFFFFF00"
            kwargs["edgecolor"] = _RIVERLANE_NEUTRAL_GREY
            kwargs["hatch"] = "///"
            kwargs["ecolor"] = _scale_lightness(_RIVERLANE_NEUTRAL_GREY, 0.8)
        stddev = contributions_stddev[i] if contributions_stddev is not None else None
        xerr_bounds = (
            ((min(stddev, abs(contribution)),), (stddev,))
            if stddev is not None
            else None
        )
        bar = ax.barh(
            bar_center,
            contribution,
            height=bar_height,
            left=offset,
            xerr=xerr_bounds,
            capsize=2,
            **kwargs,
        )
        text_colour = (
            _RIVERLANE_DARK_TEXT_COLOUR
            if _colour_brightness(kwargs["color"]) > 0.5
            else _RIVERLANE_WHITE_TEXT_COLOUR
        )
        if twidth < 0.8 * contribution:
            ax.bar_label(
                bar,
                labels=[description],
                label_type="center",
                color=text_colour,
            )
        else:
            non_fitting_texts.append(description)
            nft_centres.append(offset + contribution / 2)
            nft_bar_colours.append(kwargs["color"])
            nft_bar_centres.append(offset + contribution / 2)
        offset += contribution

    # Ensure that axes are fixed before adding texts and automatically place them.
    # Fix the aspect ratio for a better looking plot.
    ax.set_aspect((offset / ymax) * (1 / 5))
    # Fixing the Y-axis bounds
    ax.set_ybound(0, ymax)
    # Add labels that did not fit into the bar.
    texts = [
        ax.text(
            center,
            non_fitting_texts_ys[i % len(non_fitting_texts_ys)],
            text,
            horizontalalignment="center",
            color=colour,
        )
        for i, (center, text, colour) in enumerate(
            zip(nft_centres, non_fitting_texts, nft_bar_colours)
        )
    ]

    if texts:
        ret = adjust_text(
            texts,
            target_x=nft_bar_centres,
            target_y=[bar_top for _ in range(len(nft_bar_centres))],
            avoid_self=False,
            only_move=dict.fromkeys(("explode", "pull", "text", "static"), "x"),
            ax=ax,
        )
        if ret is None:
            warnings.warn(
                "Call to adjust_text failed. Labels will likely not be well positioned."
            )
        else:
            texts, _ = ret
    # Draw lines
    for target, text, colour in zip(nft_bar_centres, texts, nft_bar_colours):
        _draw_line_from_text_to_position(fig, ax, text, target, bar_top, colour, t=0)

    # Removing the ticks on the Y-axis as these have no meaning here.
    ax.set_yticks([])
    # Remove part of the black frame because it is not useful.
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    # Add axis labels
    ax.set_xlabel("Error budget, 1/Λ")

    if save_to_filename is not None:
        fig.savefig(save_to_filename)

    return fig, ax



# (c) Copyright Riverlane 2020-2025.
"""
Collection of dictionaries for different types of schedules for unrotated planar
codes.
"""

from dataclasses import dataclass
from typing import ClassVar

from deltakit_circuit._basic_types import Coord2DDelta


@dataclass
class UnrotatedPlanarCodeSchedules:
    """
    Dataclass to capture default syndrome extraction schedules for the unrotated
    planar code.
    """

    # In default_N_Z_dict,
    #     "N" means the N-shaped schedule, i.e. for that plaquette we use the
    #     following schedule:
    #             -------------
    #             |    /      |
    #             |   /_>__   |
    #             |       /   |
    #             |     ./    |
    #             -------------
    #     and "Z" means the Z-shaped schedule, i.e. for that plaquette we use the
    #     following schedule:
    #             -------------
    #             |      \    |
    #             |    _<_\   |
    #             |    \      |
    #             |     \.    |
    #             -------------

    default_N_Z_dict: ClassVar[
        dict[str, tuple[Coord2DDelta, Coord2DDelta, Coord2DDelta, Coord2DDelta]]
    ] = {
        "N": (
            Coord2DDelta(0, 1),
            Coord2DDelta(-1, 0),
            Coord2DDelta(1, 0),
            Coord2DDelta(0, -1),
        ),
        "Z": (
            Coord2DDelta(0, 1),
            Coord2DDelta(1, 0),
            Coord2DDelta(-1, 0),
            Coord2DDelta(0, -1),
        ),
    }

    # Horizontally reflected version of default_N_Z_dict, i.e. with
    #             -------------
    #             |      \    |
    #             |    _<_\   |
    #             |    \      |
    #             |     \.    |
    #             -------------
    # and
    #             -------------
    #             |    /      |
    #             |   /_>__   |
    #             |       /   |
    #             |     ./    |
    #             -------------
    horizontally_reflected_N_Z_dict: ClassVar[
        dict[str, tuple[Coord2DDelta, Coord2DDelta, Coord2DDelta, Coord2DDelta]]
    ] = {
        "N": (
            Coord2DDelta(0, 1),
            Coord2DDelta(1, 0),
            Coord2DDelta(-1, 0),
            Coord2DDelta(0, -1),
        ),
        "Z": (
            Coord2DDelta(0, 1),
            Coord2DDelta(-1, 0),
            Coord2DDelta(1, 0),
            Coord2DDelta(0, -1),
        ),
    }

    # Vertically reflected version of default_N_Z_dict, i.e. with
    #             -------------
    #             |     '\    |
    #             |    _>_\   |
    #             |    \      |
    #             |     \     |
    #             -------------
    # and
    #             -------------
    #             |    /'     |
    #             |   /_<__   |
    #             |       /   |
    #             |      /    |
    #             -------------
    vertically_reflected_N_Z_dict: ClassVar[
        dict[str, tuple[Coord2DDelta, Coord2DDelta, Coord2DDelta, Coord2DDelta]]
    ] = {
        n_or_z: schedule[::-1]
        for n_or_z, schedule in horizontally_reflected_N_Z_dict.items()
    }

    # Doubly reflected version of default_N_Z_dict, i.e. with
    #             -------------
    #             |    /'     |
    #             |   /_<__   |
    #             |       /   |
    #             |      /    |
    #             -------------
    # and
    #             -------------
    #             |     '\    |
    #             |    _>_\   |
    #             |    \      |
    #             |     \     |
    #             -------------
    doubly_reflected_N_Z_dict: ClassVar[
        dict[str, tuple[Coord2DDelta, Coord2DDelta, Coord2DDelta, Coord2DDelta]]
    ] = {n_or_z: schedule[::-1] for n_or_z, schedule in default_N_Z_dict.items()}



# (c) Copyright Riverlane 2020-2025.
"""
This module contains an implementation of IBM's Bivariate Bicycle
qLDPC codes as specified in https://arxiv.org/pdf/2308.07915.pdf.

The code is flexible and should allow exploration of codes with parameters
other than those listed in the paper.
"""
from __future__ import annotations

from functools import reduce
from itertools import product

import galois
import numpy as np
import numpy.typing as npt
from deltakit_circuit import PauliX, PauliZ, Qubit
from deltakit_circuit._basic_types import Coord2D

from deltakit_explorer.codes._css._css_code import CSSCode
from deltakit_explorer.codes._logicals import css_code_compute_logicals
from deltakit_explorer.codes._stabiliser import Stabiliser


def _find_anticommuting_pairs(
    x_logs_as_vecs: list[list[int]], z_logs_as_vecs: list[list[int]], k: int, n: int
) -> tuple[list[list[int]], list[list[int]]]:
    """
    Given a set of logical X and Z operators, find k anti-commuting pairs of
    logical operators that are also independent.

    Given a set of x and z logical operators,
    (0) for each x (z) find all z (x) that anticommute
    (1) pick an x and find a z that anticommutes with it
    (2) for each other z that anticommutes with the chosen x, multiply it by the chosen z so that it commutes with the chosen x
    (3) for each other x that anticommutes with the chosen z, multiply it by the chosen x so that it commutes with the chosen z
    Repeat until a set of desired size is found.

    Logical operators are converted into integers for optimsation purposes. This is because
    integer operations can be completed in one CPU clock cycle whereas numpy arrays can take longer.
    The integers are computed from the logical operator's binary vector representation, by just
    interpreting the vector as an integer.

    Params
    ------
    x_logs_as_vecs : List[List[int]]
        List of all logical X operators, as binary vectors, e.g, Lists with entry 0 or 1.
    z_logs_as_vecs : List[List[int]]
        List of all logical Z operators, as binary vectors, e.g, Lists with entry 0 or 1.
    k : int
        Number of logical qubits in the code.
    n : int
        Number of physical qubits in the code.

    Returns
    -------
    Tuple[List[List[int]], List[List[int]]]
        Tuple of lists of k X and Z logical operators respectively, as binary vectors,
        that is, Lists with entry 0 or 1.
    """
    # convert to ints for optimsation
    x_logs = [int("".join("01"[i] for i in vec), 2) for vec in x_logs_as_vecs]
    z_logs = [int("".join("01"[i] for i in vec), 2) for vec in z_logs_as_vecs]

    chosen_x_logs: list[int] = []
    chosen_z_logs: list[int] = []
    while len(chosen_x_logs) < k:
        if len(x_logs) == 0 or len(z_logs) == 0:
            msg = f"Unable to construct {k} logical operators, could only find {len(chosen_x_logs)}"
            raise ValueError(msg)
        # get which x anti-commutes with which z and vice-versa
        # use dictionaries for constant-time lookup
        x_ac_with_z: dict[int, dict] = {x: {} for x in x_logs}
        z_ac_with_x: dict[int, dict] = {z: {} for z in z_logs}
        for x_log, z_log in product(x_logs, z_logs):
            if (x_log & z_log).bit_count() % 2 == 1:
                x_ac_with_z[x_log][z_log] = {}
                z_ac_with_x[z_log][x_log] = {}

        # pick the first available x and first available z that
        # anti-commutes with that x
        chosen_x = x_logs[0]
        try:
            chosen_z = next(iter(x_ac_with_z[chosen_x].keys()))
        except IndexError:
            # this logical X doesnt have any anticommuting Zs, so
            # try a different X
            del x_logs[0]
            continue
        chosen_x_logs.append(chosen_x)
        chosen_z_logs.append(chosen_z)

        # update the remaining logicals by multiplying in chosen z to the other zs
        # that anti-commutes with the chosen x, and vice versa
        for z_log_i, z_log in enumerate(z_logs):
            if z_log in x_ac_with_z[chosen_x]:
                z_logs[z_log_i] = z_log ^ chosen_z
        for x_log_i, x_log in enumerate(x_logs[1:]):
            if x_logs[x_log_i + 1] in z_ac_with_x[chosen_z]:
                x_logs[x_log_i + 1] = x_log ^ chosen_x

        # remove these logicals from the running
        x_logs.remove(chosen_x)
        z_logs.remove(
            0
        )  # chosen z will be zero since it will have been xor'd with itself

    # convert from int back to binary array
    chosen_x_logs_as_vec = [
        [int(bit) for bit in f"{op_as_int:b}".zfill(n)] for op_as_int in chosen_x_logs
    ]
    chosen_z_logs_as_vec = [
        [int(bit) for bit in f"{op_as_int:b}".zfill(n)] for op_as_int in chosen_z_logs
    ]
    return chosen_x_logs_as_vec, chosen_z_logs_as_vec


class Monomial:
    """
    Class used to encapsulate the "Monomial" concept used in the
    IBM paper.

    Monomials are expressions of the form (x^a)*(y^b) where the
    powers a and b are modulo the code parameters `l` and `m` respectively.

    These monomials describe individual data qubits, and a collection
    of Monomials (Polynomials) can be used to describe a collection of data qubits.

    The inverse of a Monomial is defined to be another Monomial with powers a', b'
    such that a+a'0, b+b'=0. That is, a Monomial that one can multiply by to get back
    to the identity element x^0 y^0.

    Params
    ------
    x_pow : int
        The power of the x-term in the Monomial.
    y_pow : int
        The power of the y-term in the Monomial.
    l : int
        The l parameter of the code the Monomial is
        a part of.
    m : int
        The m parameter of the code the Monomnial is
        a part of.
    """

    def __init__(self, x_pow: int, y_pow: int, l: int, m: int):  # noqa: E741
        if l > 0 and m > 0:
            self.l = l
            self.m = m
        else:
            msg = "l and m must be >= 0"
            raise ValueError(msg)
        self.x_pow = x_pow % l
        self.y_pow = y_pow % m

    def __mul__(self, other: object) -> Monomial:
        if not isinstance(other, Monomial):
            msg = f"Can only multiply Monomials by other Monomials, not {type(other)}"
            raise NotImplementedError(msg)
        if self.l != other.l or self.m != other.m:
            msg = (
                "Cannot multiply monomials of differing max degree."
                f" lhs: l={self.l}, m={self.m} | rhs: l={other.l}, m={other.m}"
            )
            raise ValueError(msg)
        return Monomial(
            (other.x_pow + self.x_pow) % self.l,
            (other.y_pow + self.y_pow) % self.m,
            self.l,
            self.m,
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Monomial):
            return False
        return (
            self.x_pow == other.x_pow
            and self.y_pow == other.y_pow
            and self.l == other.l
            and self.m == other.m
        )

    def __repr__(self) -> str:
        return f"Monomial({self.x_pow}, {self.y_pow}, {self.l}, {self.m})"

    def __str__(self) -> str:
        return f"x^{self.x_pow} y^{self.y_pow}"

    def inverse(self) -> Monomial:
        """
        The inverse of a monomial is a monomial with powers
        such that when multiplied, the powers add to 0.
        """
        return Monomial(
            (self.l - self.x_pow) % self.l,
            (self.m - self.y_pow) % self.m,
            self.l,
            self.m,
        )

    def __rmul__(self, lhs: Monomial):
        return self.__mul__(lhs)


class Polynomial:
    """
    Class to encapsulate the "Polynomial" concept used in
    the IBM paper.
    """

    def __init__(self, monomials: list[Monomial]):
        self.monomials = monomials

    @staticmethod
    def from_vec(vec: list[int], l, m) -> Polynomial:  # noqa: E741
        """
        Create a Polynomial from a binary vector.

        Args
        ----
        vec : List[int]
            Binary vector, of the form of a List of integer entries 0 or 1.
            The vector is interpreted as the indices of the qubits that are
            represented by the Polynomial. E.g, index 0 in the vector is
            qubit 0, index 1 is qubit 1, and a 0 or 1 in these indices
            indicates whether that qubit is a part of the polynomial or not.
        l : int
            Max dimension of x component.
        m : int
            Max dimension of y component.

        Returns
        -------
        Polynomial
            Polynomial from the vector with monomial elements corresponding
            to the position of each 1 in the vector.
        """
        vec_of_monomials = []
        for i, x in enumerate(vec):
            if x != 0:
                a_i = i // m
                b_i = i % m
                vec_of_monomials.append(Monomial(a_i, b_i, l, m))
        return Polynomial(vec_of_monomials)

    def to_vec(self) -> list[int]:
        """
        Convert a Polynomial back to a binary vector.

        Returns
        -------
        List[int]
            Int (bool) vector representation of the Polynomial.
        """
        if len(self.monomials) == 0:
            return []
        vec = [0] * (self.monomials[0].l * self.monomials[0].m)
        for mon in self.monomials:
            vec[(mon.x_pow * mon.m) + mon.y_pow] = 1
        return vec

    def reverse(self) -> Polynomial:
        """
        Return reverse of a polynomial by computing the inverse
        of its constituent Monomials.
        """
        if len(self.monomials) == 0:
            return Polynomial([])
        return Polynomial([mon.inverse() for mon in self.monomials])

    def mult_by_monomial(self, monomial: Monomial) -> Polynomial:
        """
        Multiply a monomial into a polynomial. Will element-wise multiply
        the monomial into each monomial in the polynomial.

        Args
        ----
        monomial : Monomial
            Monomial to multiply into polynomial

        Returns
        -------
        Polynomial
            Result of multiplication.
        """
        if len(self.monomials) == 0:
            return Polynomial([])
        return Polynomial([mon * monomial for mon in self.monomials])

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Polynomial):
            return False
        return all(m1 == m2 for m1, m2 in zip(self.monomials, other.monomials))

    def __repr__(self) -> str:
        return f"{[str(mon) for mon in self.monomials]}"

    def __str__(self) -> str:
        return self.__repr__()


class BivariateBicycleCode(CSSCode):
    r"""
    Class to represent the IBM Bivariate Bicycle qLDPC codes, as specified
    in arXiv:2308.07915.
    If `validate=True`, will check a series of conditions as the code is
    constructed to validate that the produced code is a valid BB code.

    Parameters
    ----------
    param_l : int
        The parameter `l`, used to construct a code of length
        `n = 2lm`.
    param_m : int
        The parameter `m`, used to construct a code of length
        `n = 2lm`.
    m_A_powers : List[int]
        The powers of the polynomial terms used to construct
        the matrix A. Each polynomial is of the form
        `x^a + y^b + y^c` so the sequence should specify
        `[a, b, c]`.
    m_B_powers : List[int]
        The powers of the polynomial terms used to construct
        the matrix B. Each polynomial is of the form
        `y^a + x^b + x^c` so the sequence should specify
        `[a, b, c]`.
    validate : bool, Optional
        If True, will perform a series of check on the code's
        parameters to ensure the constructed code is a valid BB code.
        By default, True.
    check_logical_operators_are_independent : bool, optional
        Whether to check if logical X-Z operator pairs are independent. If True, then
        the commuting relations [X_k, Z_l] = 0 (k != l) are checked for the logical
        operators, ensuring that the logical operators correspond to separate logical
        qubits. If False, this check is ignored, in which case the logical operators
        can be dependent, e.g. instead of only accepting X_1, X_2 and Z_1, Z_2, the
        input X_1, X_1*X_2 and Z_1, Z_1*Z_2 is also accepted.
        Note this is forced to False if validate is False.
        By default, False.

    Attributes
    ----------
    n : int
        The parameter n as in the [[n,k,d]] specification of an error-correcting code.
        E.g, number of physical qubits.
    k : int
        The parameter n as in the [[n,k,d]] specification of an error-correcting code.
        E.g, number of logical qubits.
    m_x : npt.NDArray[np.int\_]
        A binary matrix representing the matrix `x` in the paper.
        Used to construct matrices A and B for constructing the
        parity check matrices.
    m_y : npt.NDArray[np.int\_]
        A binary matrix representing the matrix `y` in the paper.
        Used to construct matrices A and B for constructing the
        parity check matrices.
    m_A_submatrices : Tuple[npt.NDArray[np.int\_], npt.NDArray[np.int\_], npt.NDArray[np.int\_]]
        A length-3 tuple containing the submatrices A1, A2 and A3 of A.
        Used to calculate stabilisers.
    m_B_submatrices : Tuple[npt.NDArray[np.int\_], npt.NDArray[np.int\_], npt.NDArray[np.int\_]]
        A length-3 tuple containing the submatrices B1, B2 and B3 of B.
        Used to calculate stabilisers.
    m_Hx : npt.NDArray[np.int\_]
        Parity check matrix for X checks.
    m_Hz : npt.NDArray[np.int\_]
        Parity check matrix for Z checks.
    """

    def __init__(
        self,
        param_l: int,
        param_m: int,
        m_A_powers: list[int],
        m_B_powers: list[int],
        validate: bool = True,
        check_logical_operators_are_independent: bool = False,
    ):
        if validate:
            self._validate_input_parameters(param_l, param_m, m_A_powers, m_B_powers)
        self.param_l = param_l
        self.param_m = param_m
        self.m_A_powers = m_A_powers
        self.m_B_powers = m_B_powers

        # work out code parameters
        self.n = 2 * param_l * param_m

        # create S and I for l and m
        m_S_l = np.roll(np.eye(param_l, dtype=np.int_), shift=1, axis=1)
        m_S_m = np.roll(np.eye(param_m, dtype=np.int_), shift=1, axis=1)
        m_I_l = np.eye(param_l, dtype=np.int_)
        m_I_m = np.eye(param_m, dtype=np.int_)

        # create x and y. Note that x and y commute
        m_x = np.kron(m_S_l, m_I_m)
        m_y = np.kron(m_I_l, m_S_m)

        if validate:
            # assert properties of x,y to catch errors
            self._assert_x_y_properties(self.param_l, self.param_m, m_x, m_y)
        self.m_x = m_x
        self.m_y = m_y

        # create A and B
        # A = x^a + y^b + y^c, B = y^d + x^e + x^f
        # add powers of x and y to create A and B
        m_A1 = np.linalg.matrix_power(m_x, m_A_powers[0])
        m_A2 = np.linalg.matrix_power(m_y, m_A_powers[1])
        m_A3 = np.linalg.matrix_power(m_y, m_A_powers[2])
        m_A = reduce(np.add, [m_A1, m_A2, m_A3]) % 2

        m_B1 = np.linalg.matrix_power(m_y, m_B_powers[0])
        m_B2 = np.linalg.matrix_power(m_x, m_B_powers[1])
        m_B3 = np.linalg.matrix_power(m_x, m_B_powers[2])
        m_B = reduce(np.add, [m_B1, m_B2, m_B3]) % 2

        if validate:
            # assert properties of A and B:
            self._assert_matrix_ma_mb_properties(m_A, m_B)
        self.m_A_submatrices = (m_A1, m_A2, m_A3)
        self.m_B_submatrices = (m_B1, m_B2, m_B3)

        # create Hx and Hz check matrices. Note that since A and B
        # commute, these are valid check matrices
        self.m_Hx = np.hstack((m_A, m_B))
        self.m_Hz = np.hstack((m_B.T, m_A.T))

        # place qubits on a square grid
        (
            self._mat_col_to_l_data_map,
            self._mat_col_to_r_data_map,
            self._mat_col_to_x_anc_map,
            self._mat_col_to_z_anc_map,
        ) = self._get_qubit_coords()

        self._gf2 = galois.GF2
        self.k = (
            self.n
            - np.linalg.matrix_rank(self._gf2(self.m_Hx))
            - np.linalg.matrix_rank(self._gf2(self.m_Hz))
        )

        # work out stabilisers
        stabilisers = self._get_stabilisers()

        # work out logicals
        x_logicals, z_logicals = self._get_logicals()

        # construct CSSCode
        super().__init__(
            stabilisers=[stabilisers],
            x_logical_operators=x_logicals,
            z_logical_operators=z_logicals,
            use_ancilla_qubits=True,
            check_logical_operators_are_independent=check_logical_operators_are_independent,
        )

    @staticmethod
    def _validate_input_parameters(
        ell: int,
        m: int,
        m_A_powers: list[int],
        m_B_powers: list[int],
    ):
        """
        Validate that the input parameters are valid as specified by
        the IBM paper.

        Params
        ------
        ell : int
            Parameter `l` as in the paper.
        m : int
            Parameter `m` as in the paper.
        m_A_powers : List[int]
            Powers for the individual A_i matrices to be raised to.
        m_B_powers : List[int]
            Powers for the individual B_i matrices to be raised to.
        """
        if ell < 1:
            msg = "param_l should be greater than or equal to 1."
            raise ValueError(msg)
        if m < 1:
            msg = "param_m should be greater than or equal to 1."
            raise ValueError(msg)
        if len(m_A_powers) != 3:
            msg = "m_A_powers should contain 3 integers."
            raise ValueError(msg)
        if len(m_B_powers) != 3:
            msg = "m_B_powers should contain 3 integers."
            raise ValueError(msg)
        if [n == 0 for n in m_A_powers + m_B_powers].count(True) > 1:
            msg = "Should only be at most one power of 0 in A and B powers"
            raise ValueError(msg)
        for n in [m_A_powers[0], *m_B_powers[1:]]:
            if not 0 <= n < ell:
                msg = "Powers of x must be in between 0 and l-1 inclusive"
                raise ValueError(msg)
            if (m_B_powers[1] - m_B_powers[2]) % ell == 0:
                msg = "Powers of x in the same polynomial must be distinct"
                raise ValueError(msg)
        for n in [m_B_powers[0], *m_A_powers[1:]]:
            if not 0 <= n < m:
                msg = "Powers of y must be in between 0 and m-1 inclusive"
                raise ValueError(msg)
            if (m_A_powers[1] - m_A_powers[2]) % m == 0:
                msg = "Powers of y in the same polynomial must be distinct"
                raise ValueError(msg)

    @staticmethod
    def _assert_x_y_properties(
        param_l: int, param_m: int, m_x: npt.NDArray, m_y: npt.NDArray
    ):
        # (1) xy = yx
        if not np.allclose(m_x @ m_y, m_y @ m_x):
            msg = "y*x should equal x*y"
            raise ValueError(msg)

        # (2) x^l = y^m = I_lm
        if not np.allclose(
            np.linalg.matrix_power(m_x, param_l),
            np.identity(param_l * param_m, dtype=np.int_),
        ):
            msg = "x^l should equal the identity matrix of shape (l*m, l*m)"
            raise ValueError(msg)
        if not np.allclose(
            np.linalg.matrix_power(m_y, param_m),
            np.identity(param_l * param_m, dtype=np.int_),
        ):
            msg = "y^m should equal the identity matrix of shape (l*m, l*m)"
            raise ValueError(msg)

    @staticmethod
    def _assert_matrix_ma_mb_properties(m_A: npt.NDArray, m_B: npt.NDArray):
        # (1) exactly 3 non-zero entries in any row or column
        if not (np.count_nonzero(m_A, axis=0) == 3).all():
            msg = "Matrix A should have exactly 3 non-zero entries in each column"
            raise ValueError(msg)
        if not (np.count_nonzero(m_A, axis=1) == 3).all():
            msg = "Matrix A should have exactly 3 non-zero entries in each row"
            raise ValueError(msg)
        if not (np.count_nonzero(m_B, axis=0) == 3).all():
            msg = "Matrix B should have exactly 3 non-zero entries in each column"
            raise ValueError(msg)
        if not (np.count_nonzero(m_B, axis=1) == 3).all():
            msg = "Matrix B should have exactly 3 non-zero entries in each row"
            raise ValueError(msg)

        # (2) AB = BA
        if not np.allclose(m_A @ m_B, m_B @ m_A):
            msg = "A*B should equal B*A"
            raise ValueError(msg)

    def _get_qubit_coords(
        self,
    ) -> tuple[dict[int, Qubit], dict[int, Qubit], dict[int, Qubit], dict[int, Qubit]]:
        """
        Attempt to lay out the qubits on a square-ish grid, the size of which is dictated
        by the periodic boundary conditions of the torus the code may or may not
        naturally lie on. Not all BB codes permit a toric layout; see Lemma 4 of the
        paper for more details. Hence, if the code is one of codes listed in Table 3 of
        the paper, this code will work, but is not guaranteed to work (and likely won't)
        for codes outside of these.

        We arbitrarily pick the first row of the Hx matrix to be the bottom-left-most
        X ancilla qubit. We may then follow the scheduling of the syndrome measurement to
        determine the placement of the data qubits and rest of the ancillas.

        The method used is to start in the bottom left of the matrix, calling this
        Qubit(0,0) arbitrarily, then move along the rows assigning Qubits(1,0),
        ubits(2,0) etc as per the construction described in the paper, detailed above.
        There are rules there for how the submatrices connect to each other which dictate
        what the local connections are and what the long range connections are.

        NB: The [[90, 8, 10]] code is a special case and uses different parameters, but
        should still arrive at a similar grid layout.

        Returns
        -------
        Tuple[Dict[int, Qubit], Dict[int, Qubit], Dict[int, Qubit], Dict[int, Qubit]]
            A tuple of dictionaries mapping the index of rows/columns of the parity
            check matrices to qubits, in order:
                left data qubits, right data qubits, x ancillas, z ancillas.
            For instance, column 0 of the Hx (WLOG Hz) matrix is the first data qubit in
            the "L" (R) group of data qubits, and would be described in dictionary 0 (1)
            with entry {0: Qubit(...)}. The first row of Hx (Hz) is the first X (Z)
            ancilla and is described by dictionary 2 (3) with entry {0: Qubit(...)}.
        """
        # assign the first row of the Hx matrix to be the bottom-left-most qubit
        x_coord, y_coord, matrix_index_of_next_qubit = 0, 0, 0
        x_lookup: dict[int, Qubit] = {}
        z_lookup: dict[int, Qubit] = {}
        dl_lookup: dict[int, Qubit] = {}
        dr_lookup: dict[int, Qubit] = {}

        # when scanning over qubit connections, the bottom row alternates Z-anc to L-data
        # and alternate rows between R-data and X-anc, so use the qubit coord mod 2 to
        # access the correct dictionary - y % 2 is which type of row we are, x % 2 is whether
        # we are assigning a data or an ancilla
        dict_lookup = ((z_lookup, dl_lookup), (dr_lookup, x_lookup))

        # when scanning over rows, the submatrix we must check to find the adjacent qubit
        # changes based on the type of qubit we are currently looking at.
        # keep the relevant dictionaries in tuples so that they may be accessed using the
        # current coordinate mod 2. for a deeper explanation of the choice of matrices here,
        # refer to Fig 4 of the paper, along with Lemma 3 and 4.
        submatrix_lookup_y = (
            # z_anc -> R_d, L_d -> x_anc
            (self.m_A_submatrices[1].T, self.m_A_submatrices[1].T),
            # R_d -> z_anc, x_anc -> l_d
            (self.m_A_submatrices[2], self.m_A_submatrices[2]),
        )
        if (
            self.param_l == 15
            and self.param_m == 3
            and self.m_A_powers == [9, 1, 2]
            and self.m_B_powers == [0, 2, 7]
        ):
            submatrix_lookup_x = (
                # z_anc -> L_d, L_d -> z_anc
                (self.m_B_submatrices[2].T, self.m_B_submatrices[0]),
                # R_d -> x_anc, x_anc -> R_d
                (self.m_B_submatrices[2].T, self.m_B_submatrices[0]),
            )
        else:
            submatrix_lookup_x = (
                (self.m_B_submatrices[2].T, self.m_B_submatrices[1]),
                (self.m_B_submatrices[2].T, self.m_B_submatrices[1]),
            )
        max_iters = 6 * self.n  # should only need 2*n
        current_iter_index = 0
        while current_iter_index < max_iters:
            dict_lookup[y_coord % 2][x_coord % 2][matrix_index_of_next_qubit] = Qubit(
                Coord2D(x_coord, y_coord)
            )
            matrix_index_of_next_qubit = np.where(
                submatrix_lookup_x[y_coord % 2][x_coord % 2][matrix_index_of_next_qubit]
                == 1
            )[0][0]
            x_coord += 1
            if matrix_index_of_next_qubit in dict_lookup[y_coord % 2][x_coord % 2]:
                # reached the torus boundary, so move up a row via the relevant connection
                matrix_index_of_next_qubit = np.where(
                    submatrix_lookup_y[y_coord % 2][x_coord % 2][
                        matrix_index_of_next_qubit
                    ]
                    == 1
                )[0][0]
                x_coord = 0
                y_coord += 1
                if matrix_index_of_next_qubit in dict_lookup[y_coord % 2][x_coord % 2]:
                    # reached torus y boundary, so we're done
                    break
            current_iter_index += 1
        else:
            msg = (
                "Max iteration limit exceeded for coord placement; consider"
                " changing code parameters"
            )
            raise ValueError(msg)

        return dl_lookup, dr_lookup, x_lookup, z_lookup

    def _get_stabilisers(self) -> list[Stabiliser]:
        """
        Given the appropriate matrices describing the code, construct
        the stabilisers with valid scheduling.

        Returns
        -------
        List[Stabiliser]
            List of alternating X and Z stabilisers. Starting with X.
        """
        # use the order in equation (9), Section 5
        # this order was found empirically and has no easy intuitive explanation.
        # essentially, we just pick one check at a time from each A or B sub matrix
        # and order them such that they are hopefully distance preserving.
        zip_x_stab_qubit_coords = [
            # 2 indicates left qubits, which for X type is A
            # 3 is then right which is B
            (self._mat_col_to_l_data_map, self.m_A_submatrices[1]),  # A2
            (self._mat_col_to_r_data_map, self.m_B_submatrices[1]),  # B2
            (self._mat_col_to_r_data_map, self.m_B_submatrices[0]),  # B1
            (self._mat_col_to_r_data_map, self.m_B_submatrices[2]),  # B3
            (self._mat_col_to_l_data_map, self.m_A_submatrices[0]),  # A1
            (self._mat_col_to_l_data_map, self.m_A_submatrices[2]),  # A3
        ]
        zip_z_stab_qubit_coords = [
            # 2 indicates left qubits, which for Z type is B
            # 3 is then right which is A
            (self._mat_col_to_r_data_map, self.m_A_submatrices[0]),  # A1
            (self._mat_col_to_r_data_map, self.m_A_submatrices[2]),  # A3
            (self._mat_col_to_l_data_map, self.m_B_submatrices[0]),  # B1
            (self._mat_col_to_l_data_map, self.m_B_submatrices[1]),  # B2
            (self._mat_col_to_l_data_map, self.m_B_submatrices[2]),  # B3
            (self._mat_col_to_r_data_map, self.m_A_submatrices[1]),  # A2
        ]
        x_stabilisers = []
        z_stabilisers = []
        for check_index in range(self.param_l * self.param_m):
            # X stabilisers
            x_stabilisers.append(
                Stabiliser(
                    paulis=[
                        None,
                        *(
                            PauliX(
                                col_ind_to_qubit_dict[
                                    np.nonzero(matrix[check_index, :])[0][0]
                                ]
                            )
                            for col_ind_to_qubit_dict, matrix in zip_x_stab_qubit_coords
                        ),
                    ],
                    ancilla_qubit=self._mat_col_to_x_anc_map[check_index],
                )
            )

            # Z stabilisers
            z_stabilisers.append(
                Stabiliser(
                    paulis=[
                        *(
                            PauliZ(
                                col_ind_to_qubit_dict[
                                    np.nonzero(matrix[:, check_index])[0][0]
                                ]
                                # indices of non-zero elements in row of matrix
                            )
                            for col_ind_to_qubit_dict, matrix in zip_z_stab_qubit_coords
                        ),
                        None,
                    ],
                    ancilla_qubit=self._mat_col_to_z_anc_map[check_index],
                )
            )

        return x_stabilisers + z_stabilisers

    def _get_logicals(
        self,
    ) -> tuple[list[list[PauliX[Coord2D]]], list[list[PauliZ[Coord2D]]]]:
        """
        Will return a tuple of lists of the X and Z logicals, respectively.
        They are paired up in order, that is, they anticommute when the
        indices of the two lists are aligned and commute otherwise.

        Returns
        -------
        Tuple[List[List[PauliX[Coord2D]]], List[List[PauliZ[Coord2D]]]]
            Tuple of X and Z logical operators, in order such that
            x_logicals[i],z_logicals[j] anti-commute exactly when i=j
            and commute otherwise.
        """
        # get a subset of logicals from css_code, and compute the rest
        x_logs, _ = css_code_compute_logicals(self.m_Hx, self.m_Hz)

        # get f and g,h from X logicals
        # X logicals are of the form X(f,0) and X(g,h), eq. (16)
        f_vecs, g_vecs, h_vecs = [], [], []
        for x_log in x_logs:
            x_log_list = x_log.reshape(-1).tolist()
            half = len(x_log_list) // 2
            if all(x == 0 for x in x_log_list[half:]):
                f_vecs.append(x_log_list[:half])
            else:
                g_vecs.append(x_log_list[:half])
                h_vecs.append(x_log_list[half:])

        # logical operators are cyclic (as are stabilisers) so we
        # may permute them by monomials.
        # we have a subset of all possible f,g,h, so find the rest
        # by multiplying in each monomial to get a set of |M|=lm
        all_monomials_of_lm = [
            Monomial(x, y, self.param_l, self.param_m)
            for x, y in product(range(self.param_l), range(self.param_m))
        ]
        all_f_vecs = f_vecs.copy()
        all_gT_vecs = [
            Polynomial.from_vec(vec, self.param_l, self.param_m).reverse().to_vec()
            for vec in g_vecs
        ]
        all_hT_vecs = [
            Polynomial.from_vec(vec, self.param_l, self.param_m).reverse().to_vec()
            for vec in h_vecs
        ]
        for list_to_extend, list_to_iterate, is_g_or_h in [
            (all_f_vecs, f_vecs, False),
            (all_gT_vecs, g_vecs, True),
            (all_hT_vecs, h_vecs, True),
        ]:
            for vec in list_to_iterate:
                vec_as_poly = Polynomial.from_vec(vec, self.param_l, self.param_m)
                for mon in all_monomials_of_lm:
                    if is_g_or_h:
                        shifted_vec = (
                            vec_as_poly.reverse().mult_by_monomial(mon).to_vec()
                        )
                    else:
                        shifted_vec = vec_as_poly.mult_by_monomial(mon).to_vec()
                    if not np.any(np.equal(list_to_extend, shifted_vec).all(1)):
                        list_to_extend.append(shifted_vec)

        # now combine f,g,h to make logical operators
        # create the unprimed block, X(f, 0) and Z(h.T, g.T)
        all_f_logs = [list(f) + [0] * (self.n // 2) for f in all_f_vecs]
        all_gh_logs = [list(hT) + list(gT) for gT, hT in zip(all_gT_vecs, all_hT_vecs)]

        # find pairs with correct commutation relations
        # find anticommuting pairs of logical operators
        unprime_x_logs, unprime_z_logs = _find_anticommuting_pairs(
            all_f_logs, all_gh_logs, self.k // 2, self.n
        )

        # get primed block from unprimed
        # X(g, h) and Z(0, f.T)
        prime_x_logs, prime_z_logs = [], []
        for up_x_log in unprime_x_logs:
            prime_z_logs.append(
                [0] * (len(up_x_log) // 2)
                + Polynomial.from_vec(
                    up_x_log[: len(up_x_log) // 2], self.param_l, self.param_m
                )
                .reverse()
                .to_vec()
            )
        for up_z_log in unprime_z_logs:
            prime_x_logs.append(
                Polynomial.from_vec(
                    up_z_log[(len(up_z_log) // 2) :], self.param_l, self.param_m
                )
                .reverse()
                .to_vec()
                + Polynomial.from_vec(
                    up_z_log[: (len(up_z_log) // 2)], self.param_l, self.param_m
                )
                .reverse()
                .to_vec()
            )

        # convert binary arrays to arrays of Pauli terms.
        # the first coordinate is whether its in A or B submatrix
        # the second coordinate is the relative coordinate within A or B
        dict_lookup = {2: self._mat_col_to_l_data_map, 3: self._mat_col_to_r_data_map}
        x_logicals_as_paulis = [
            [
                PauliX(
                    dict_lookup[2 + j // (self.param_l * self.param_m)][
                        j % (self.param_l * self.param_m)
                    ]
                )
                for j in np.where(logical)[0]
            ]
            for logical in unprime_x_logs + prime_x_logs
        ]
        z_logicals_as_paulis = [
            [
                PauliZ(
                    dict_lookup[2 + j // (self.param_l * self.param_m)][
                        j % (self.param_l * self.param_m)
                    ]
                )
                for j in np.where(logical)[0]
            ]
            for logical in unprime_z_logs + prime_z_logs
        ]

        return x_logicals_as_paulis, z_logicals_as_paulis



# (c) Copyright Riverlane 2020-2025.
"""
* `NoiseParameters` treats the quantum system in terms of one- and two-qubit
    noise, which comes together with gate application, measurement, reset,
    or intervals of idleness.
"""

from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass, field
from itertools import chain

import numpy as np
from deltakit_circuit import NoiseProfile, Qubit
from deltakit_circuit.gates import _MeasurementGate
from deltakit_circuit.noise_channels import (
    Depolarise1,
    NoiseChannel,
    OneQubitNoiseChannel,
    PauliChannel1,
)


def _idle_noise_from_t1_t2(
    t1: float, t2: float
) -> Callable[[Qubit, float], OneQubitNoiseChannel]:
    """
    Return a function that calculates idle noise as a function of T1 and T2
    times. See `Ghosh et al. <https://arxiv.org/abs/1210.5799>`_ for the
    derivation, with the final result in equation (10).

    Parameters
    ----------
    t1 : float
        T1 time in s.
    t2 : float
        T2 time in s.

    Returns
    -------
    Callable[[Qubit, float], OneQubitNoiseChannel]
        Function which takes a qubit and a time in s as inputs and returns
        a one-qubit noise channel.
    """
    if t1 <= 0.0:
        msg = "Relaxation time `t1` must be positive."
        raise ValueError(msg)
    if t2 <= 0.0:
        msg = "Dephasing time `t2` must be positive."
        raise ValueError(msg)
    if t2 >= 2 * t1:
        msg = "Dephasing time `t2` must be less than twice relaxation time `t1`."
        raise ValueError(msg)

    def _get_idle_noise(qubit: Qubit, t: float) -> OneQubitNoiseChannel:
        if t1 == t2:
            return Depolarise1(qubit, 0.75 * (1.0 - np.exp(-t / t1)))
        return PauliChannel1(
            qubit,
            0.25 * (1.0 - np.exp(-t / t1)),
            0.25 * (1.0 - np.exp(-t / t1)),
            0.5 * (1.0 - np.exp(-t / t2)) - 0.25 * (1.0 - np.exp(-t / t1)),
        )

    return _get_idle_noise


@dataclass(kw_only=True)
class NoiseParameters:
    """
    Dataclass to capture a noise model for a quantum computer.

    Parameters
    ----------
    gate_noise : Dict[Type[Gate], Callable[[Gate], NoiseChannel]]
        Noise to be added after applying a gate. This is specified through a
        dictionary with gate classes as keys and functions which take a gate and
        return a noise channel as values. The noise can thus be different for each
        gate class and can be dependent on the qubit(s) involved in the gate. By
        default, there is no noise.

    idle_noise : Optional[Callable[[Qubit, float], NoiseChannel]]
        Noise to be added to an idle qubit. This can be dependent on the qubit and the
        time for which it is idle. By default, there is no noise.

    reset_noise : Optional[Callable[[OneQubitResetGate], NoiseChannel]]
        Noise to be added after resetting a qubit. This can be dependent on the
        reset gate applied and the qubit involved. By default, there is no noise.

    measurement_noise_after : Optional[Callable[[OneQubitMeasurementGate], NoiseChannel]]
        Noise to be added after measuring a qubit. This can be dependent on the
        measurement gate applied and the qubit involved. By default, there is no noise.

    measurement_flip : Optional[Callable[[_MeasurementGate], _MeasurementGate]]
        Mapping from a measurement gate to a noisy measurement gate, capturing adding
        error associated with obtaining an incorrect measurement result. By default,
        there is no noise.

    measurement_noise_before : Optional[Callable[[OneQubitMeasurementGate], NoiseChannel]]
        Noise which will be applied before a measurement gate. By default, there is no noise.

    Raises
    ------
    ValueError
        If any key in gate_noise is not a one-qubit or two-qubit gate.
    """

    name = "noise_model"

    gate_noise: list[NoiseProfile] = field(default_factory=list)
    idle_noise: Callable[[Qubit, float], NoiseChannel] | None = None
    reset_noise: list[NoiseProfile] = field(default_factory=list)
    measurement_noise_after: list[NoiseProfile] = field(default_factory=list)
    measurement_noise_before: list[NoiseProfile] = field(default_factory=list)

    measurement_flip: dict[
        type[_MeasurementGate], Callable[[_MeasurementGate], _MeasurementGate]
    ] = field(default_factory=dict)

    def as_noise_profile_after_gate(self) -> list[NoiseProfile]:
        """
        Returns the noise profiles encapsulated by this object as a single generator
        of noise profiles. Omits the noise before measurement.
        """
        return list(
            chain(
                self.gate_noise,
                self.reset_noise,
                self.measurement_noise_after,
            )
        )



# (c) Copyright Riverlane 2020-2025.
from deltakit_explorer.qpu._native_gate_set import (
                                                    ExhaustiveGateSet,
                                                    NativeGateSet,
                                                    NativeGateSetAndTimes,
)
from deltakit_explorer.qpu._noise._noise_parameters import NoiseParameters
from deltakit_explorer.qpu._noise._phenomenological_noise import PhenomenologicalNoise
from deltakit_explorer.qpu._noise._physical_noise import PhysicalNoise
from deltakit_explorer.qpu._noise._sd6_noise_model import SD6Noise
from deltakit_explorer.qpu._noise._si1000_noise import SI1000Noise
from deltakit_explorer.qpu._noise._toy_noise import ToyNoise
from deltakit_explorer.qpu._qpu import QPU

# List only public members in `__all__`.
__all__ = [s for s in dir() if not s.startswith("_")]



# (c) Copyright Riverlane 2020-2025.
"""
The `Logging` class is a service class which enables keeping track of the calls
made and errors which occurred. In the log file, `./client.log`, there will be a
trace of all calls and exceptions that occurred during your sessions. Pay attention to
the `reqid=[...]` fields in this file. Each request to the server is accompanied by
a unique string, which may be used to facilitate incident investigation.
"""


from __future__ import annotations

import inspect
import logging
import logging.handlers
import uuid
from typing import Any

from deltakit_explorer._utils._utils import get_log_directory

LOG_FILENAME = get_log_directory() / "client.log"
LOG_FORMAT = "[%(asctime)s.%(msecs)03d][%(name)s][%(levelname)s] %(message)s"


class Logging:
    """This class enables client-side logging, generates unique request identifiers
    and proxies events from the calling environment (e.g. JupyterHub or gql library).
    """

    # inherit logging values
    CRITICAL = logging.CRITICAL
    FATAL = logging.FATAL
    ERROR = logging.ERROR
    WARNING = logging.WARNING
    INFO = logging.INFO
    DEBUG = logging.DEBUG
    NOTSET = logging.NOTSET

    # switch off requests logging, as it logs full data strings
    logging.getLogger("requests").disabled = True
    logging.getLogger("urllib3").disabled = True
    # logs in a file will be searchable
    # with cat client.log | grep "\[deltakit-explorer\]"
    logger = logging.getLogger("deltakit-explorer")
    logger.setLevel(logging.WARNING)
    file_handler = logging.handlers.RotatingFileHandler(
        LOG_FILENAME, maxBytes=10 * 1024 * 1024, backupCount=3
    )
    file_handler.setFormatter(logging.Formatter(LOG_FORMAT))
    # log to file
    logger.addHandler(file_handler)
    # log to stderr stream by default
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(logging.Formatter(LOG_FORMAT))
    logger.addHandler(stream_handler)

    @staticmethod
    def info_and_generate_uid(args: dict[str, Any]) -> str:
        """This method is used to log the API method invocation,
        and to generate a unique string which represents the
        call in the log files on both client and server sides.

        Args:
            args (Dict[str, Any]):
                any arguments, assumed to be a result
                of `locals()` call inside the method. Will be saved
                as is, with `self` value removed.

        Returns:
            str: unique string, identifying function call.
        """
        # get a calling function name
        name = inspect.stack()[1].function
        # generate a unique ID
        uid = uuid.uuid4().hex
        args.pop("self", None)

        newargs = {}
        for key, value in args.items():
            # long string or any other object representation
            string_value = str(value)
            if len(string_value) >= 64:
                string_value = string_value[:30] + "..." + string_value[-30:]
            newargs[key] = string_value
        Logging.logger.info("%s(%s); reqid=[%s]", name, newargs, uid)
        return uid

    @staticmethod
    def info(message: str, uid: str):
        """This method is used to log events during the API method
        invocation. This method should be used after `info_and_generate_uid()`,
        and requires a `uid` as an argument::

            uid = Logging.info_and_generate_uid(locals())
            ...
            Logging.info("Your message", uid)

        Args:
            message (str): message
            uid (str): unique string, result of `info_and_generate_uid()`
        """
        name = inspect.stack()[1].function
        Logging.logger.info("%s: %s | reqid=[%s]", name, message, uid)

    @staticmethod
    def warn(message: str, uid: str):
        """This method is used to log warnings during the API method
        invocation. This method should be used after `info_and_generate_uid()`,
        and requires a `uid` as an argument::

            uid = Logging.info_and_generate_uid(locals())
            ...
            Logging.warn("Your message", uid)

        Args:
            message (str): warning message
            uid (str): unique string, result of `info_and_generate_uid()`
        """
        name = inspect.stack()[1].function
        Logging.logger.warning("%s: %s | reqid=[%s]", name, message, uid)

    @staticmethod
    def error(ex: Exception, uid: str):
        """Use this call for `Exception` logging.
        `uid` obtained from `info_and_generate_uid()` is required::

            uid = Logging.info_and_generate_uid(locals())
            ...
            try:
                # raises here
            except Exception as e:
                Logging.error(e, uid)

        Args:
            ex (Exception): exception object from `except Exception as ex:`.
            uid (str): unique string, result of `info_and_generate_uid()`
        """
        name = inspect.stack()[1].function
        Logging.logger.error("%s: %s | reqid=[%s]", name, ex, uid)

    @classmethod
    def set_log_to_console(cls, log_to_console: bool):
        """Switches console logging on and off."""
        stream_handlers = [
            handler
            for handler in cls.logger.handlers
            # check exactly console logger
            if type(handler)  # pylint: disable=unidiomatic-typecheck
            is logging.StreamHandler
        ]
        if log_to_console:
            if len(stream_handlers) == 0:
                cls.logger.addHandler(cls.stream_handler)
        else:
            # remove all stream handlers
            for handler in stream_handlers:
                cls.logger.removeHandler(handler)

    @classmethod
    def set_log_level(cls, loglevel: int):
        """
        Sets logging level, use `logging.INFO`,
        `logging.DEBUG`, ... values.
        """
        cls.logger.setLevel(loglevel)



# (c) Copyright Riverlane 2020-2025.
"""`analysis` module aggregates client-side analytical tools."""

from __future__ import annotations

from collections.abc import Sequence

import numpy as np
import numpy.typing as npt
from deltakit_core.deprecation import deprecated
from packaging.version import Version


@deprecated(
    reason="A better function is now available.",
    replaced_by="deltakit.explorer.analysis.compute_logical_error_per_round",
    removed_in_version=Version("1.0.0"),
)
def get_exp_fit(
    logical_fails_all_rounds: npt.NDArray[np.int_] | list[int],
    shots_all_rounds: npt.NDArray[np.int_] | list[int],
    all_rounds: list[int],
    interpolation_points: int = 26,
) -> tuple[
    float,
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
]:
    """
    Implement logical error probability per round fit as described in
    https://arxiv.org/pdf/2310.05900.pdf (p.40) and
    https://arxiv.org/pdf/2207.06431.pdf (p.21). The first round (`r=0`)
    data points are excluded as the error suppression is stronger there
    than in consecutive rounds.

    Args:
        logical_fails_all_rounds (npt.NDArray | List[int]):
            Number of logical failures per round of experiment.
        shots_all_rounds (npt.NDArray | List[int]):
            Number of shots per round of experiment.
        all_rounds (List[int]):
            The number of rounds corresponding to each data
            point, e.g. [1, 2, 4, 8, 16].
        interpolation_points: int
            How many points to use in the interpolated plot.
            Default is 26.

    Returns:
        Tuple[float, npt.NDArray, npt.NDArray, npt.NDArray]:
            A tuple consisting of
            - an epsilon parameter value (error probability per round);
            - an interpolated graph (2 value), computed with this value;
            - error bars value.

    Examples:
        Calling the method::

            eps, x, y, yerr = Analysis.get_exp_fit(
                logical_fails_all_rounds=[20, 20, 30, 40, 50],
                shots_all_rounds=[500, 500, 500, 500, 500],
                all_rounds=[1, 3, 5, 7, 9],
                interpolation_points=25 + 1,
            )

    """
    # Calculate logical fidelity (F = 1 - 2 * p_err) and its standard error
    logical_perr_per_round = np.array(logical_fails_all_rounds) / np.array(
        shots_all_rounds
    )
    fidelity = 1 - 2 * logical_perr_per_round
    assert not np.any(fidelity < 0.0), (
        f"Fidelity values (1-2*p) should be non-negative, but were {fidelity}."
    )

    yerr = np.sqrt(
        logical_perr_per_round * (1 - logical_perr_per_round) / shots_all_rounds
    )

    # Take base-10 log of fidelity and find
    # the best linear fit
    sigma = np.abs(yerr / fidelity)[1:]
    # Where `sigma = 0`, a weight will be infinite. This will cause
    # a warning and error in `polyfit`. The error is enough, so let's
    # suppress the warnings. If the error is considered confusing, we
    # can catch and re-raise with more information.
    with np.errstate(divide="ignore"):
        w = 1 / sigma
    x = np.array(all_rounds[1:])
    y = np.log10(fidelity[1:])

    # cov=True may return the covariance matrix
    with np.errstate(invalid="ignore"):
        pf = np.polyfit(
            x=x,
            y=y,
            deg=1,
            w=w,
        )

    # Exponentiate curve fit params to obtain decay rate
    # epsilon and initial fidelity f_0
    epsilon = float(0.5 * (1.0 - 10.0 ** pf[0]))
    f_0 = float(10.0 ** pf[1])

    # Interpolate x values across the range of available rounds
    # and calculate y values based on an exponential decay with
    # rate epsilon and initial fidelity f_0.
    rounds_interpolated = np.linspace(
        all_rounds[0],
        all_rounds[-1],
        interpolation_points,
        dtype=np.float64,
    )
    y_interpolated = [f_0 * (1 - 2 * epsilon) ** r for r in rounds_interpolated]
    probs_interpolated = (1.0 - np.array(y_interpolated, dtype=np.float64)) * 0.5

    return epsilon, rounds_interpolated, probs_interpolated, yerr


def calculate_lep_and_lep_stddev(
    fails: npt.NDArray[np.int_] | Sequence[int] | int,
    shots: npt.NDArray[np.int_] | Sequence[int] | int,
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """Calculate the logical error probability and its standard deviation.

    Args:
        fails (npt.NDArray[np.int\\_] | Sequence[int] | int):
            The number of logical failures.
        shots (npt.NDArray[np.int\\_] | Sequence[int] | int):
            The number of shots the experiment was run for.

    Returns:
        tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
            A tuple consisting of
            - the logical error probability;
            - the standard deviation of the logical error probability.

    Examples:
        Calculating logical error probability and standard deviation
        given number of fails, and number of shots::

            lep, lep_stddev = Analysis.calculate_lep_and_lep_stddev(
                fails=[498, 151, 34],
                shots=[500000] * 3,
            )
    """
    fails, shots = np.asarray(fails), np.asarray(shots)
    lep = fails / shots
    if np.any(lep <= 0):
        msg = "Must have > 0 fails to calculate logical error probability."
        raise ValueError(msg)
    lep_stddev = np.sqrt(lep * (1 - lep) / shots)

    return lep, lep_stddev



@deprecated(
    reason=(
        "A better function is now available. The output of this function can now be "
        "computed with error bars and any number of points."
    ),
    replaced_by="deltakit.explorer.analysis.calculate_lambda_and_lambda_stddev",
    removed_in_version=Version("1.0.0"),
)
def get_lambda_fit(
    distances: list[int], lep_per_round: list[float], lep_stddev_per_round: list[float]
) -> npt.NDArray[np.float64]:
    """
    Get the best fit line with gradient lambda for plotting purposes.

    Accepts the logical error probability (LEP) per round, which may be approximated
    as LEP / num_rounds (for small LEP), and equally for lep_stddev.

    Args:
        distances (List[int]):
            The distances of the code.
        lep_per_round (List[float]):
            The logical error probabilities per round.
        lep_stddev_per_round (List[float]):
            The standard deviation of the logical error probabilities per round.

    Returns:
        npt.NDArray:
            The best fit line of log(lep_per_round) vs distance, from which
            the gradient is lambda.

    Examples:
        Fit exponential curve given logical error probability for 5, 7, and
        9 rounds of QEC experiment::

            lep_per_round_fit = Analysis.get_lambda_fit(
                distances=[5, 7, 9],
                lep_per_round=[1.992e-04, 4.314e-05, 7.556e-06],
                lep_stddev_per_round=[1.2e-05, 9.3e-06, 3.9e-06],
            )

    """
    params, _ = np.polyfit(
        x=distances,
        y=np.log(lep_per_round),
        deg=1,
        w=[
            lep_pr / lep_std_pr
            for lep_pr, lep_std_pr in zip(lep_per_round, lep_stddev_per_round)
        ],
        cov="unscaled",
    )

    return np.exp(params[0] * np.array(distances) + params[1])



# (c) Copyright Riverlane 2020-2025.
"""
This module aggregates functions to manipulate `deltakit_circuit.Circuit` objects.
These include native gate compilation, noise introduction,
and various optimisations.
This module is not currently public; this __init__.py file is a historical artefact
and can be removed, adjusting imports within other `deltakit_explorer` modules
accordingly.
"""


from deltakit_explorer.qpu._circuits._circuit_functions import (
    merge_layers,
    remove_identities,
)
from deltakit_explorer.qpu._circuits._parallelise_circuits import (
    parallelise_disjoint_circuits,
    parallelise_same_length_circuits,
)
from deltakit_explorer.qpu._circuits._tableau_compile_functions import (
    _compile_measurement_to_native_gates_plus_unitaries,
    _compile_or_exchange_unitary_block,
    _compile_reset_and_meas_to_native_gates,
    _compile_reset_to_native_gates_plus_unitaries,
    _compile_two_qubit_gate_to_target,
    _compile_two_qubit_gates_to_native_gates,
    compile_circuit_to_native_gates,
)
from deltakit_explorer.qpu._circuits._tableau_functions import (
    CZ_TO_GATE_DICT,
    CZSWAP_TO_GATE_DICT,
    GATE_TO_CZ_DICT,
    GATE_TO_CZSWAP_DICT,
    MEAS_COMPILATION_LOOKUP_DICT,
    RESET_COMPILATION_LOOKUP_DICT,
    CompilationData,
    _create_circuit_from_compilation_data,
    _extract_structure_from_circuit,
    _get_compilation_dict,
    _get_compilation_with_measurement_after_unitaries,
    _get_compilation_with_projectors_before_unitaries,
    _get_compilation_with_two_qubit_gates,
    _is_identity_like,
)

__all__ = [s for s in dir() if not s.startswith("_")]



# (c) Copyright Riverlane 2020-2025.
"""Description of ``deltakit.explorer.data`` namespace here."""

from deltakit_explorer.data._converter import (
                                               read_01,
                                               read_b8,
                                               read_csv,
                                               write_01,
                                               write_b8,
                                               write_binary_data_to_file,
)

# List only public members in `__all__`
__all__ = [s for s in dir() if not s.startswith("_")]



# (c) Copyright Riverlane 2020-2025.
"""This file contains types to support QEC experiments.
"""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any

import stim
from deltakit_circuit.gates import PauliBasis

from deltakit_explorer.enums._api_enums import (
    DataFormat,
    QECECodeType,
    QECExperimentType,
)
from deltakit_explorer.types._types import (
    BinaryDataType,
    CircuitParameters,
    DetectionEvents,
    JSONable,
    LeakageFlags,
    Measurements,
    ObservableFlips,
)


@dataclass
class QECExperiment:
    """
    Represents a core of simulated experiment:
     - a noisy circuit which is used for simulation;
     - sweep bits, if used for initialisation;
     - obtained measurements, detector, observable and leakage values.
    """

    noisy_circuit: str
    sweep_bits: BinaryDataType | None = None

    measurements: Measurements | None = None

    detectors: DetectionEvents | None = None
    observables: ObservableFlips | None = None
    leakage_flags: LeakageFlags | None = None

    @staticmethod
    def from_circuit_and_measurements(
        stim_path: Path,
        measurements_path: Path,
        measurements_format: DataFormat,
        sweep_path: Path | None = None,
        sweep_format: DataFormat | None = None,
    ) -> QECExperiment:
        """Build an experiment object from 2 files.

        Args:
            stim_path (Path): location of a circuit.
            measurements_path (Path):
                location of a measurements file.
            measurements_format (DataFormat):
                format of measurement data.
            sweep_path (Optional[Path]):
                location of sweep bits, optional.
            sweep_format (Optional[DataFormat]):
                format of a sweep bits file.

        Returns:
            QECExperiment: experiment object.

        Raises:
            ValueError:
                If only sweep_path or sweep_format provided.
        """
        if (sweep_format is None) ^ (sweep_path is None):
            msg = (
                "sweep_path and sweep_format should be both provided, "
                "or should both be None."
            )
            raise ValueError(
                msg
            )
        circuit = stim.Circuit.from_file(stim_path)
        measurements = Measurements(
            measurements_path,
            measurements_format,
            circuit.num_measurements,
        )
        sweeps = None
        if sweep_path and sweep_format:
            sweeps = BinaryDataType(
                sweep_path,
                sweep_format,
                circuit.num_sweep_bits,
            )
        return QECExperiment(
            noisy_circuit=str(circuit),
            measurements=measurements,
            sweep_bits=sweeps,
        )

    def compute_detectors_and_observables(self):
        """If measurements and a circuit are provided, ensure that
        detectors and observables are computed."""
        if (
            (self.detectors is None or self.observables is None)
            and self.measurements is not None
        ):
            dets, obs = (
                self.measurements.to_detectors_and_observables(
                    stim_circuit=self.noisy_circuit,
                    sweep_bits=self.sweep_bits,
                )
            )
            if self.detectors is None:
                self.detectors = dets
            if self.observables is None:
                self.observables = obs


@dataclass
class QECExperimentDefinition(JSONable):
    """Definition, essential to generate a QEC experiment circuit."""

    experiment_type: QECExperimentType
    """Type of experiment, e.g. Quantum Memory or Stability."""

    code_type: QECECodeType
    """QEC code, e.g. Rotated Planar Code."""

    observable_basis: PauliBasis
    """Basis, in which observable is measured. E.g. X or Z."""

    num_rounds: int
    """Number of experiment rounds."""

    basis_gates: list[str] | None = None
    """If circuit is generated with a specific basis gate set."""

    parameters: CircuitParameters | None = None
    """Parameters of circuit generation."""

    def get_parameters_gql_string(self) -> dict[str, Any] | None:
        """
        Prepare graphql parameters to be inserted
        into GraphQL query.
        """
        if self.parameters is not None:
            return self.parameters.to_gql()
        return None

    @staticmethod
    def get_repetition_z_quantum_memory(
        distance: int,
        num_rounds: int,
        basis_gates: list[str] | None = None
    ) -> QECExperimentDefinition:
        """Convenience method to quickly define a Z-memory
        repetition code experiment."""
        return QECExperimentDefinition(
            experiment_type=QECExperimentType.QUANTUM_MEMORY,
            code_type=QECECodeType.REPETITION,
            observable_basis=PauliBasis.Z,
            num_rounds=num_rounds,
            basis_gates=basis_gates,
            parameters=CircuitParameters.from_sizes([distance]),
        )

    @staticmethod
    def get_rotated_planar_z_quantum_memory(
        distance: int,
        num_rounds: int,
        basis_gates: list[str] | None = None
    ) -> QECExperimentDefinition:
        """Convenience method to quickly define a Z-memory
        (d x d)-rotated planar code experiment."""
        return QECExperimentDefinition(
            experiment_type=QECExperimentType.QUANTUM_MEMORY,
            code_type=QECECodeType.ROTATED_PLANAR,
            observable_basis=PauliBasis.Z,
            num_rounds=num_rounds,
            basis_gates=basis_gates,
            parameters=CircuitParameters.from_sizes([distance, distance]),
        )

    def to_json(self):
        dump = super().to_json()
        dump["generate_circuit_parameters"] = dump.pop("parameters")
        dump["qe_code_type"] = dump.pop("code_type")
        dump["rounds"] = dump.pop("num_rounds")
        dump["basis"] = dump.pop("observable_basis")
        dump["basis_gate_set"] = dump.pop("basis_gates")
        return dump



# (c) Copyright Riverlane 2020-2025.

"""
This module contains a class representing a stabiliser of a stabiliser code.
"""

from __future__ import annotations

from collections.abc import Collection

from deltakit_circuit import Qubit
from deltakit_circuit._qubit_identifiers import PauliGate, T


class Stabiliser:
    """
    Class representing a stabiliser.

    Parameters
    ----------
    paulis: Collection[PauliGate | None]
        The elementary Pauli terms in a stabiliser. If the input is ordered,
        this order specifies the schedule for syndrome extraction and
        None indicates an empty layer in the syndrome extraction circuit.
        A basic syndrome extraction circuit, using RX, CX, CY, CZ and MX
        gates, is constructed in the Stage class. For instance, `[PauliX(0),
        None, PauliY(1)]` corresponds to a syndrome extraction circuit with layers:

        - RX ancilla,
        - CX ancilla --> data qubit 0,
        - empty layer,
        - CY ancilla --> data qubit 1,
        - MX ancilla.

        Compilation to other gates is done in QPU class.
    ancilla_qubit: Optional[Qubit[T] | T]
        Ancilla qubit or None. Can be specified either directly as a Qubit or as an
        identifier (e.g., int). In the None case, we implicitly mean that
        the object doesn't have an associated syndrome extraction circuit.
        By default, it is None.

    Attributes
    ----------
    data_qubits: Set[Qubit]
        Set of all data qubits on which the stabiliser is defined.
    """

    def __init__(
        self,
        paulis: Collection[PauliGate | None],
        ancilla_qubit: Qubit[T] | T | None = None,
    ):
        """
        Raises
        ------
        ValueError
            If the ancilla qubit also acts as a data qubit and the stabiliser
            operator consists of more than one Pauli.
        """
        if isinstance(ancilla_qubit, Qubit):
            self.ancilla_qubit: Qubit[T] = ancilla_qubit
        elif ancilla_qubit is None:
            self.ancilla_qubit = ancilla_qubit
        else:
            self.ancilla_qubit = Qubit(ancilla_qubit)

        self._check_data_qubits([pauli.qubit for pauli in paulis if pauli is not None])
        self.data_qubits: set[Qubit] = {
            pauli.qubit for pauli in paulis if pauli is not None
        }

        if self.ancilla_qubit in self.data_qubits and len(paulis) > 1:
            # For now we accept the ancilla being the data_qubit, if we have only
            # one pauli in paulis. This is useful when putting together stages into
            # an experiment.
            msg = "Ancilla qubit should be different from the data qubits."
            raise ValueError(msg)

        self.paulis: tuple[PauliGate | None, ...] = tuple(paulis)

    @staticmethod
    def _check_data_qubits(qubits: Collection[Qubit]) -> None:
        """
        Check qubits contains only unique elements and is not empty.

        Raises
        ------
        ValueError
            If qubits contains duplicate elements.
        ValueError
            If qubits is empty.
        """
        if len(set(qubits)) == 0:
            msg = "Stabiliser was initialised without Pauli terms."
            raise ValueError(msg)
        if len(set(qubits)) != len(qubits):
            msg = "Data qubits given in paulis should be unique."
            raise ValueError(msg)

    @property
    def operator_repr(self) -> set[PauliGate]:
        """
        A set of PauliX/Y/Z's representing the Pauli operator. Useful
        when comparing Stabilisers.

        Returns
        -------
        Set[PauliGate]
            Pauli operator representation.
        """
        return {pauli for pauli in self.paulis if pauli is not None}

    def __eq__(self, other: object) -> bool:
        """
        Check if two Stabiliser objects have the same paulis and
        ancilla_qubit attributes.
        """
        return (
            isinstance(other, Stabiliser)
            and self.paulis == other.paulis
            and self.ancilla_qubit == other.ancilla_qubit
        )

    def __hash__(self) -> int:
        return hash((self.__class__, self.paulis, self.ancilla_qubit))

    def __mul__(self, other: Stabiliser) -> Stabiliser:
        """
        Multiply two Pauli products with disjoint supports, as operators. The
        returned object's ancilla_qubit is None!
        E.g. X_1 X_2 * Z_3 Y_4 will be X_1 X_2 Z_3 Y_4, but X_1 X_2 * Z_1 is not
        defined!

        Returns
        -------
        Stabiliser
            The product without ancilla.
        """
        if isinstance(other, Stabiliser) and self.data_qubits.isdisjoint(
            other.data_qubits
        ):
            return Stabiliser(
                paulis=[
                    pauli for pauli in self.paulis + other.paulis if pauli is not None
                ],
            )
        return NotImplemented

    def __repr__(self) -> str:
        return f"Stabiliser({self.paulis}, {self.ancilla_qubit})"



# (c) Copyright Riverlane 2020-2025.
"""Description of ``deltakit.explorer.simulation`` namespace here."""

from deltakit_explorer.simulation._stim_simulation import simulate_with_stim

# List only public members in `__all__`.
__all__ = [s for s in dir() if not s.startswith("_")]



import itertools
from collections.abc import Callable, Mapping, Sequence
from concurrent.futures import ProcessPoolExecutor

import numpy as np
import numpy.typing as npt
from deltakit_circuit._circuit import Circuit
from deltakit_decode._mwpm_decoder import PyMatchingDecoder
from deltakit_decode.analysis._matching_decoder_managers import StimDecoderManager
from tqdm import tqdm

from deltakit_explorer.analysis.error_budget._memory import (
    MemoryGenerator,
    get_rotated_surface_code_memory_circuit,
)


def _generate_surface_code_memory_decoder_manager(
    distance: int,
    num_rounds: int,
    noise_model: Callable[[Circuit, npt.NDArray[np.floating]], Circuit],
    noise_parameters: npt.NDArray[np.floating],
    memory_generator: MemoryGenerator,
    noise_parameter_names: Sequence[str],
) -> StimDecoderManager:
    """Generate a decoder manager with a rotated planar code memory experiment.

    Args:
        distance (int): distance of the rotated planar code memory experiment that
            should be generated by this function.
        num_rounds (int): number of stabilizer measurement rounds of the memory
            experiment generated by this function.
        noise_model (Callable[[Circuit, npt.NDArray[np.floating]], Circuit]): a callable
            adding noise to the provided circuit, according to the parameters provided
            in the second argument.
        noise_parameters: parameters for the ``noise_model`` callable representing the
            noise strength.
        memory_generator: a callable that is able to generate memory experiments from a
            given distance and number of rounds.
        noise_parameter_names: if provided, human-readable names for each of the
            provided ``noise_parameters``.

    Returns:
        a decoder manager that can then be run using ``RunAllAnalysisEngine``.
    """
    circuit = memory_generator(distance, num_rounds)
    noisy_circuit = noise_model(circuit, noise_parameters)
    decoder, decoder_circuit = PyMatchingDecoder.construct_decoder_and_stim_circuit(
        noisy_circuit
    )
    metadata = {
        "distance": distance,
        "num_rounds": num_rounds,
        **{
            f"noise_{name}": p
            for name, p in zip(noise_parameter_names, noise_parameters, strict=True)
        },
    }

    return StimDecoderManager(decoder_circuit, decoder, metadata=metadata)


def _generate_surface_code_memory_decoder_manager_wrapper(
    data: tuple[
        int,
        int,
        Callable[[Circuit, npt.NDArray[np.floating]], Circuit],
        npt.NDArray[np.floating],
        MemoryGenerator,
        Sequence[str],
    ],
) -> StimDecoderManager:
    return _generate_surface_code_memory_decoder_manager(*data)


def generate_decoder_managers_for_lambda(
    xi: npt.NDArray[np.floating],
    noise_model: Callable[[Circuit, npt.NDArray[np.floating]], Circuit],
    num_rounds_by_distances: Mapping[int, Sequence[int]],
    max_workers: int = 1,
    memory_generator: MemoryGenerator = get_rotated_surface_code_memory_circuit,
    noise_parameter_names: Sequence[str] | None = None,
) -> list[StimDecoderManager]:
    """Generate several decoder managers from the provided arguments for a rotated
    planar code memory experiment.

    This function should be used to generate in one call a description of all the
    sampling tasks that should be done in order to estimate the value of Λ on the
    provided noise parameters ``xi``.

    Note:
        Generating the tasks might be a long process, hence the presence of the
        ``max_worker`` parameter.

    Warning:
        This function will generate ``len(num_rounds) * len(distances) *
        numpy.prod(xi.shape[1:])`` sampling tasks. That quantity
        might grow quickly, so try to avoid providing too many entries in ``distances``
        and ``num_rounds`` except if you really need it.

    Args:
        xi (npt.NDArray[numpy.floating]): an array of shape ``(m, n)``, where ``m`` is
            the number of noise parameters needed by ``noise_model`` and ``n`` the
            number of points to evaluate Λ on.
        noise_model (Callable[[Circuit, npt.NDArray[np.floating]], Circuit]): a callable
            adding noise to the provided circuit, according to the parameters provided.
        num_rounds_by_distances (Mapping[int, Sequence[int]]): the different number of
            rounds that will be used to estimate the logical error-rate per round for
            each code distance. As a rule of thumb, at least ``4`` entries should be
            provided and the maximum number of rounds should, if possible, give a
            logical error-rate above ``0.2`` for the largest distance and a logical
            error-rate below ``0.45`` for the smallest distance. You will not be able to
            check those thresholds everytime, but the standard errors computed will be
            lower if they are verified.
        max_workers (int): if provided, the number of parallel processes that are used
            to generate the sampling tasks. If not, default to ``1``.
        memory_generator: a callable that is able to generate memory experiments from a
            given distance and number of rounds.
        noise_parameter_names: if provided, human-readable names for each of the
            provided ``noise_parameters``. Defaults to the noise parameter index (i.e.,
            "0", "1", ...).

    Returns:
        a list of decoder managers that can then be run using ``RunAllAnalysisEngine``.
    """
    if noise_parameter_names is None:
        noise_parameter_names = [str(i) for i in range(xi.shape[0])]
    # 0. Transposing to be able to iterate on different values.
    noise_parameters = xi.T
    # 1. Check input parameters
    if max_workers <= 0:
        msg = f"Cannot have less than one worker. Asked for {max_workers} workers."
        raise ValueError(msg)
    # 2. Build an iterator on all the parameters we need to provide to the generation
    # method.
    total_circuits = (
        sum(len(num_rounds) for num_rounds in num_rounds_by_distances.values())
        * noise_parameters.shape[0]
    )
    # (distance: int, num_rounds: int)
    distance_and_rounds_iterator = itertools.chain.from_iterable(
        zip(itertools.repeat(d, len(rounds)), rounds, strict=True)
        for d, rounds in num_rounds_by_distances.items()
    )
    parameters_iterator = (
        (
            distance,
            num_rounds,
            noise_model,
            noise_parameter,
            memory_generator,
            noise_parameter_names,
        )
        for ((distance, num_rounds), noise_parameter) in itertools.product(
            distance_and_rounds_iterator, noise_parameters
        )
    )
    # 3. Generate the decoder managers
    decoder_managers: list[StimDecoderManager] = []
    if max_workers == 1:
        for parameters in tqdm(
            parameters_iterator,
            total=total_circuits,
            desc="Generating quantum circuits",
        ):
            decoder_managers.append(
                _generate_surface_code_memory_decoder_manager(*parameters)
            )
        return decoder_managers
    # Else, we might use several workers
    with ProcessPoolExecutor(max_workers=max_workers) as pool:
        return list(
            tqdm(
                iterable=pool.map(
                    _generate_surface_code_memory_decoder_manager_wrapper,
                    parameters_iterator,
                ),
                total=total_circuits,
                desc="Generating quantum circuits",
            )
        )



import warnings
from collections.abc import Callable, Sequence
from dataclasses import dataclass
from math import floor

import numpy as np
import numpy.typing as npt
from scipy.optimize import curve_fit


@dataclass(frozen=True)
class LogicalErrorProbabilityPerRoundResults:
    """Named-tuple-like class containing computation results from
    :func:`compute_logical_error_per_round`.

    Attributes:
        leppr (float): Logical Error Probability Per Round (LEPPR).
        leppr_stddev (float): LEPPR standard deviation.
        spam_error (float): computed SPAM error probability.
        spam_error_stddev (float): SPAM error probability standard deviation.
    """

    leppr: float
    leppr_stddev: float
    spam_error: float
    spam_error_stddev: float


def compute_logical_error_per_round(
    num_rounds: npt.NDArray[np.int_] | Sequence[int],
    logical_error_probabilities: npt.NDArray[np.floating] | Sequence[float],
    logical_error_probabilities_stddev: npt.NDArray[np.floating] | Sequence[float],
    *,
    force_include_single_round: bool = False,
) -> LogicalErrorProbabilityPerRoundResults:
    """Compute the logical error probability per round from different logical error
    probability computations.

    This function implements the method described in:

    1. https://arxiv.org/pdf/2310.05900.pdf (p.40)
    2. https://arxiv.org/pdf/2207.06431.pdf (p.21)
    3. https://arxiv.org/pdf/2505.09684.pdf (p.8)

    to recover an estimator of the logical error probability per round from the
    estimated values of logical error probabilities for several round durations.

    Args:
        num_rounds (npt.NDArray[numpy.int_] | Sequence[int]):
            a sequence of integers representing the number of rounds used to get the
            corresponding results in ``logical_error_probabilities`` and
            ``logical_error_probabilities_stddev``. Any value below 1 (``< 1``) is
            automatically removed from this list along with the corresponding values in
            ``logical_error_probabilities`` and ``logical_error_probabilities_stddev``.
            Any value equal to 1 is removed from this list along with the corresponding
            values in ``logical_error_probabilities`` and
            ``logical_error_probabilities_stddev`` iff ``force_include_single_round`` is
            ``False``. If only one data-point is provided (or left after the removal
            process described just before), the SPAM error is assumed to be ``0`` and an
            estimation will still be returned.

            Heuristically, to increase the returned estimation precision, you should try
            to provide data for rounds such that the estimated logical error probability
            for the number of rounds ``max(num_rounds)`` is approximately ``0.4``. This
            ``0.4`` value has been set to reduce fitting errors.
        logical_error_probabilities (npt.NDArray[numpy.floating] | Sequence[float]):
            logical error probabilities computed for each of the provided
            ``num_rounds``. Should be the same length as ``num_rounds``.
        logical_error_probabilities_stddev (npt.NDArray[numpy.floating] | Sequence[float]):
            standard deviation of the logical error probabilities provided in
            ``logical_error_probabilities``. Should be the same length as
            ``num_rounds``.
        force_include_single_round (bool):
            if ``True``, data obtained from 1-round experiment will be used in the
            computation if provided in ``num_rounds``. Default to ``False`` which
            results in 1-round data being ignored due to boundary effects that affect
            the final estimation. See https://arxiv.org/pdf/2207.06431.pdf (p.21).

    Returns:
        LEPPRResults: detailed results of the computation.

    Examples:
        Calculating per-round logical error probability and its standard deviation
        given number of fails, and number of shots for several rounds::

            res = compute_logical_error_per_round(
                num_failed_shots=[34, 151, 356],
                num_shots=[500000] * 3,
                num_rounds=[2, 4, 6],
            )
            leppr, leppr_stddev = res.leppr, res.leppr_stddev
            spam, spam_stddev = res.spam_error, res.spam_error_stddev

    """
    # Get the inputs as numpy arrays.
    # Sanitisation: also make sure that the inputs are sorted.
    isort = np.argsort(num_rounds)
    num_rounds = np.asarray(num_rounds)[isort]
    logical_error_probabilities = np.asarray(logical_error_probabilities)[isort]
    logical_error_probabilities_stddev = np.asarray(
        logical_error_probabilities_stddev
    )[isort]

    # Check that we do not have duplicate data for the same number of rounds as that
    # will confuse the numerical methods used in this function.
    unique_counts = np.unique_counts(num_rounds)
    non_unique_entries_mask = unique_counts.counts > 1
    if np.any(non_unique_entries_mask):
        non_unique_values = unique_counts.values[non_unique_entries_mask].tolist()
        msg = (
            "Multiple entries were provided for the following number of rounds: "
            f"{non_unique_values}. This is not supported. Please make sure you only "
            "provide one entry per number of rounds."
        )
        raise RuntimeError(msg)

    # Check that we do not have any num_rounds <= 0 entry.
    while num_rounds.size > 0 and num_rounds[0] <= 0:
        warnings.warn(
            f"Found an invalid number of rounds: {num_rounds[0]}. Number of rounds "
            "should be >= 1."
        )
        num_rounds = num_rounds[1:]
        logical_error_probabilities = logical_error_probabilities[1:]
        logical_error_probabilities_stddev = logical_error_probabilities_stddev[1:]

    # Filter out the r == 1 input if not forced to include it by the user.
    if num_rounds.size > 0 and num_rounds[0] == 1 and not force_include_single_round:
        num_rounds = num_rounds[1:]
        logical_error_probabilities = logical_error_probabilities[1:]
        logical_error_probabilities_stddev = logical_error_probabilities_stddev[1:]

    # Filter out logical error probabilities above 0.5 as that will lead to negative
    # fidelities.
    invalid_lep_indices = logical_error_probabilities > 0.5
    if np.any(invalid_lep_indices):
        warnings.warn(
            "Found at least one invalid (i.e., > 0.5) logical error probability. "
            "Ignoring all the provided logical error probabilities above 0.5."
        )
        valid_lep_indices = np.logical_not(invalid_lep_indices)
        num_rounds = num_rounds[valid_lep_indices]
        logical_error_probabilities = logical_error_probabilities[valid_lep_indices]
        logical_error_probabilities_stddev = logical_error_probabilities_stddev[
            valid_lep_indices
        ]

    # Checking the validity of the filtered data.
    if num_rounds.size == 0:
        msg = (
            "No valid data was provided. Please ensure that the data provided is "
            "correct. If you provided data, look at the warnings to understand why it "
            "was considered invalid and ignored by this function."
        )
        raise ValueError(msg)

    # If the user only provided one data point, we can use a direct approximate formula
    # without having to call a fitting function.
    if logical_error_probabilities.size == 1:
        warnings.warn(
            "Only one valid data-point provided for logical error probability per "
            "round. Continuing computation assuming that SPAM error is negligible."
        )
        rounds = num_rounds[0]
        lep = float(logical_error_probabilities[0])
        lep_stddev = float(logical_error_probabilities_stddev[0])
        # Implement Eq. (4) from section A.2.2. at page 40 of
        # https://arxiv.org/pdf/2310.05900.
        estimated_logical_error_per_round = (1 - (1 - 2 * lep) ** (1 / rounds)) / 2
        estimated_logical_error_per_round_stddev = (
            lep_stddev * (1 - 2 * lep) ** (1 / rounds - 1) / rounds
        )
        return LogicalErrorProbabilityPerRoundResults(
            estimated_logical_error_per_round,
            estimated_logical_error_per_round_stddev,
            0,
            0,
        )

    # Check if the heuristic guideline on the number of rounds is verified.
    max_logical_error_probability = np.max(logical_error_probabilities)
    if max_logical_error_probability < 0.2:
        warnings.warn(
            "The maximum estimated logical error probability "
            f"({max_logical_error_probability}) is below 0.2. The returned estimation "
            "might be better if you add data with more rounds such that the maximum "
            "estimated logical error probability is closer to 0.4."
        )

    fidelities = 1 - 2 * logical_error_probabilities
    # We want to do a linear regression on the log values of fidelity, and obtain the
    # per-round error probability like that.
    # Applying the logarithm function will change non-uniformly the standard deviation
    # of each variable, which makes the standard linear regression estimator biased. The
    # best linear unbiased estimator in that case is obtained by solving a weighted
    # least square problem where the weights corresponds to the reciprocal of the
    # variance of each observation.
    # See https://en.wikipedia.org/wiki/Weighted_least_squares.
    logfidelity = np.log(fidelities)
    # We approximate the standard deviation with an error propagation analysis. This
    # method has been tested against scipy and returns similar results.
    logfidelities_stddev = 2 * logical_error_probabilities_stddev / fidelities

    # Note that the covariance matrix is used later to estimate the logical error
    # probability per round standard deviation.
    (slope, offset), cov = curve_fit(
        lambda x, s, o: s * x + o,
        num_rounds,
        logfidelity,
        sigma=logfidelities_stddev,
        absolute_sigma=True,
        # If the error probabilities are exactly 0, the solution should be (0, 0).
        # Because we expect the error probabilities to be close to 0, start from (0, 0)
        # as a first estimate.
        p0=(0, 0),
        # Both slope and offset are used to recover a probability with the expression
        # p = (1 - numpy.exp(value)) / 2. Because a probability needs to be in [0, 1], we
        # have that value <= numpy.log(1).
        # Note: even though the below bounds are valid, setting bounds changes the
        # default optimisation method from "lm" to "trf". There is at least one
        # real-world example where setting those bounds led to incorrect results, so not
        # including them for the moment.
        # bounds=((-numpy.inf, -numpy.inf), (numpy.log(1), numpy.log(1))),
    )

    estimated_logical_error_per_round = float((1 - np.exp(slope)) / 2)
    # Compute the standard R2 (Coefficient of determination) using the formula
    # ``R2 = 1 - SSE / SST`` where SSE is the Sum of Squares Error and SST is the Sum of
    # Square Total that are computed below.
    sse = np.sum((logfidelity - offset - slope * num_rounds) ** 2)
    sst = np.sum((logfidelity - np.mean(logfidelity)) ** 2)
    r2 = float(1 - sse / sst)
    if abs(r2) < 0.98:
        warnings.warn(
            f"Got a R2 value of {r2} < 0.98. Estimation might be imprecise. Increasing "
            "the number of shots or re-performing the computation might help in removing "
            "this warning."
        )

    # Following https://arxiv.org/pdf/2505.09684v1 (Methods - Extracting logical error
    # per cycle, page 8) we estimate the variance on the logical error probability per
    # round (named Perrc below) using the formula
    #      sigma(Perrc) = (1 - Perrc) * sigma(slope)
    # The standard deviation on the linear fit parameters can be obtained through the
    # covariance matrix diagonal entries.
    slope_stddev, offset_stddev = np.sqrt(np.diagonal(cov))
    estimated_logical_error_per_round_stddev = float(
        (1 - 2 * estimated_logical_error_per_round) * slope_stddev / 2
    )
    estimated_spam_error = float((1 - np.exp(offset)) / 2)
    estimated_spam_error_stddev = float(
        (1 - 2 * estimated_spam_error) * offset_stddev / 2
    )
    return LogicalErrorProbabilityPerRoundResults(
        estimated_logical_error_per_round,
        estimated_logical_error_per_round_stddev,
        estimated_spam_error,
        estimated_spam_error_stddev,
    )


def simulate_different_round_numbers_for_lep_per_round_estimation(
    simulator: Callable[[int], tuple[int, int]],
    initial_round_number: int = 2,
    next_round_number_func: Callable[[int], int] = lambda x: 2 * x,
    maximum_round_number: int | None = None,
    heuristic_logical_error_lower_bound: float = 0.25,
    heuristic_logical_error_upper_bound: float = 0.45,
) -> tuple[npt.NDArray[np.int_], npt.NDArray[np.int_], npt.NDArray[np.int_]]:
    """Compute QEC results to estimate the logical error probability per round.

    This function aims at encapsulating the practical knowledge about logical error
    probability per round computation to help any user computing the required logical
    error probabilities for useful number of rounds.

    It repeatedly calls ``simulator`` with a number of rounds growing according to
    ``next_round_number_func``, starting from ``initial_round_number``,
    until the logical error probability is above
    ``heuristic_logical_error_lower_bound``. If the final step returned a logical error
    probability above ``heuristic_logical_error_upper_bound``, the algorithm then goes
    backward and replaces that last value with the first one under that limit.

    Args:
        simulator (Callable[[int], tuple[int, int]]):
            a callable that returns a tuple ``(num_fails, num_shots)`` from a number of
            rounds given as input.
        initial_round_number (int): initial value for the geometric series that will be
            used to generate the number of rounds.
        next_round_number_func (Callable[[int], int]): function used to compute the
            next round number that should be tested. Default to a linear scaling up to
            500 rounds and then an exponential scaling. The initial linear scaling is to
            avoid the nearby points generated at the beginning of the exponential
            scaling whereas the final exponential scaling is to avoid spending too much
            time if the noise is really low.
        maximum_round_number (int): if set, this function will stop once the next
            number of rounds (computed with ``next_round_number_func``) is above that
            threshold. If not set, only the other stopping criterions apply.
        heuristic_logical_error_lower_bound (float): minimum target logical error
            probability for the final round. Might not be verified by the return of this
            function if ``maximum_round_number`` is set and reached before that minimum
            threshold.
        heuristic_logical_error_upper_bound (float): maximal target logical error
            probability for the final round. Should be set sufficiently below ``0.5``
            such that the uncertainties (mostly due to finite sampling) on the computed
            logical error probability (LEP) are low enough to not introduce a plateau in
            the log-plot of the fidelity log(F) = log(1 - 2*LEP). Experimentally,
            ``0.45`` seems to check that.

    Returns:
        tuple[npt.NDArray[numpy.int_], npt.NDArray[numpy.int_], npt.NDArray[numpy.int_]]:
            A tuple consisting of
            - the different number of rounds corresponding to the two other entries,
            - the number of failed shots for the corresponding number of rounds,
            - the total number of shots for the corresponding number of rounds.

    Examples:
        Calculating per-round logical error probability and its standard deviation
        given number of fails, and number of shots for several rounds::

            def perfect_simulator(num_rounds: int) -> tuple[int, int]:
                error_per_round: float = 0.001
                total_error: float = (1 - error_per_round) ** num_rounds
                num_shots: int = 100_000
                num_fails = total_error * num_shots
                return num_fails, num_shots


            nrounds, nfails, nshots = (
                simulate_different_round_numbers_for_lep_per_round_estimation(
                    simulator=perfect_simulator,
                    initial_round_number=2,
                    geometric_factor=1.7,
                )
            )
    """
    if maximum_round_number is None:
        maximum_round_number = 2**30

    nrounds: list[int] = [initial_round_number]
    nfails: list[int] = []
    nshots: list[int] = []

    nfail, nshot = simulator(nrounds[-1])
    nfails.append(nfail)
    nshots.append(nshot)

    # Generate experiments until the number of repetitions is large enough (which is
    # heuristically determined as
    # ``logical error probability > heuristic_logical_error_lower_bound``).
    while (nfails[-1] / nshots[-1]) < heuristic_logical_error_lower_bound:
        new_round_number = next_round_number_func(nrounds[-1])
        if new_round_number > maximum_round_number:
            break
        nrounds.append(new_round_number)
        nfail, nshot = simulator(nrounds[-1])
        nfails.append(nfail)
        nshots.append(nshot)

    # We do not want to include logical error probabilities above
    # ``heuristic_logical_error_upper_bound``.
    # We go back using smaller steps until we find a last point that is over
    # ``heuristic_logical_error_lower_bound`` but under
    # ``heuristic_logical_error_upper_bound``.
    maximum_number_of_backward_steps: int = 5
    backward_arithmetic_factor: int = floor(
        (nrounds[-1] - nrounds[-2]) / (maximum_number_of_backward_steps + 1)
    )
    while (nfails[-1] / nshots[-1]) > heuristic_logical_error_upper_bound:
        out_of_bound_round_value = nrounds[-1]
        nrounds, nfails, nshots = nrounds[:-1], nfails[:-1], nshots[:-1]
        nrounds.append(out_of_bound_round_value - backward_arithmetic_factor)
        nfail, nshot = simulator(nrounds[-1])
        nfails.append(nfail)
        nshots.append(nshot)

    return np.asarray(nrounds), np.asarray(nfails), np.asarray(nshots)



# (c) Copyright Riverlane 2020-2025.
from __future__ import annotations

import importlib

from deltakit_explorer import __version__


def test_version():
    assert importlib.metadata.version('deltakit_explorer') == __version__



from collections.abc import Sequence

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
from deltakit_core.plotting.colours import RIVERLANE_PLOT_COLOURS
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from deltakit_explorer.analysis._lambda import LambdaResults


def _lambda_interpolated(
    lambda0: float, lambda_: float, distances: npt.NDArray[np.int_]
) -> npt.NDArray[np.floating]:
    """Computes logical error probability per round that would be obtained with the
    provided values.

    Uses the formula ``ε = 1 / Λ_0 * Λ**(-(d + 1) / 2)`` where:

    - ``ε`` is the logical error probability per round,
    - ``Λ_0`` is a multiplicative constant,
    - ``Λ`` is the error suppression factor,
    - ``d`` is the distance of the code,

    to estimate the logical error probability per round from the provided ``lambda_``
    and ``lambda0`` on the provided list of ``distances``.
    """
    return lambda_**(-(distances + 1) / 2) / lambda0


def plot_lambda(
    lambda_data: LambdaResults,
    distances: npt.NDArray[np.int_] | Sequence[int],
    lep_per_round: npt.NDArray[np.float64] | Sequence[float],
    lep_per_round_stddev: npt.NDArray[np.float64] | Sequence[float] | None = None,
    *,
    num_sigmas: int = 3,
    fig: Figure | None = None,
    ax: Axes | None = None,
) -> tuple[Figure, Axes]:
    """Plot Λ-fitting data.

    This function plots both the logical error-probability per round that has been used
    to compute Λ, the associated error-rates if provided, and the resulting fit, showing
    how close the fit is from actual data.

    Args:
        distances (npt.NDArray[numpy.int\\_] | Sequence[int]): The distances of the code.
        lep_per_round (npt.NDArray[numpy.float64] | Sequence[float]):
            The logical error probabilities per round.
        lep_stddev_per_round (npt.NDArray[numpy.float64] | Sequence[float]):
            The standard deviation of the logical error probabilities per round.

    Returns:
        tuple[Figure, Axes]: The matplotlib Figure and Axes objects containing the plot.

    Example:
        fig, ax = plot_lambda(
            distances = [5, 7, 9],
            lep_per_round = [0.15, 0.1, 0.05],
            lep_stddev_per_round = [0.01, 0.008, 0.005],
        )
        ax.set_yscale("log")
        plt.show()
    """
    if (fig is None) ^ (ax is None):
        msg = "The 'fig' and 'ax' parameters should either be both None or both set."
        raise ValueError(msg)

    if fig is None and ax is None:
        fig, ax = plt.subplots()

    # These should be already checked by the above code, but type checkers are not able
    # to infer that information, so including the asserts explicitly for type checkers
    # to understand.
    assert ax is not None
    assert fig is not None

    lengths = {len(distances), len(lep_per_round)}
    if lep_per_round_stddev is not None:
        lengths.add(len(lep_per_round_stddev))
    if len(lengths) > 1:
        msg = (
            "The lengths of 'distances', 'lep_per_round' and 'lep_per_round_stddev' "
            f"must be the same. Got the following lengths: {lengths}."
        )
        raise ValueError(msg)

    isort = np.argsort(distances)
    distances = np.asarray(distances)[isort]
    lep_per_round = np.asarray(lep_per_round)[isort]
    if lep_per_round_stddev is not None:
        lep_per_round_stddev = num_sigmas * np.asarray(lep_per_round_stddev)[isort]

    # Plot the logical error probabilities per round
    ax.errorbar(
        distances,
        lep_per_round,
        yerr=lep_per_round_stddev,
        fmt=".",
        color=RIVERLANE_PLOT_COLOURS[1],
        label=f"Logical error probabilities per round (±{num_sigmas}σ)"  # noqa: RUF001
    )

    # Plot the fitted lambda curve
    lambda_, lambda_stddev = lambda_data.lambda_, lambda_data.lambda_stddev
    lambda0, lambda0_stddev = lambda_data.lambda0, lambda_data.lambda0_stddev
    distances_interpolated = np.linspace(distances[0], distances[-1], 200)
    lambda_interpolated = _lambda_interpolated(lambda0, lambda_, distances_interpolated)
    ax.plot(
        distances_interpolated,
        lambda_interpolated,
        label=f"Fit, Λ={lambda_:.4f} ± {num_sigmas * lambda_stddev:.4f} ({num_sigmas}σ)",  # noqa: RUF001
        color=RIVERLANE_PLOT_COLOURS[1]
    )

    # Add error band to lambda curve
    lambda_interpolated_low = _lambda_interpolated(
        lambda0 - num_sigmas * lambda0_stddev,
        lambda_ - num_sigmas * lambda_stddev,
        distances_interpolated
    )
    lambda_interpolated_high = _lambda_interpolated(
        lambda0 + num_sigmas * lambda0_stddev,
        lambda_ + num_sigmas * lambda_stddev,
        distances_interpolated
    )
    ax.fill_between(
        distances_interpolated,
        lambda_interpolated_low,
        lambda_interpolated_high,
        color=RIVERLANE_PLOT_COLOURS[0],
        alpha=0.2
    )

    ax.set_title("Logical Error Probability Per Round Fit")
    ax.set_xlabel("Code distance")
    ax.set_ylabel("Error suppression factor Λ")
    ax.legend()
    return fig, ax
